"""
Module defining models for representing UI component
configurations.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Generic, Literal, TypeVar

from pathlib import Path

from pydantic import (
    BaseModel,
    ConfigDict,
    PrivateAttr,
    SerializerFunctionWrapHandler,
    model_serializer,
)

import gradio as gr
from gradio.components import Component

from ultimate_rvc.core.exceptions import ComponentNotInstatiatedError
from ultimate_rvc.typing_extra import DeviceType
from ultimate_rvc.web.typing_extra import (
    AnyCallable,
    BaseDropdownValue,
    BaseRadioValue,
    DropdownChoices,
    RadioChoices,
)

if TYPE_CHECKING:
    from collections.abc import Sequence

U = TypeVar("U")
T = TypeVar("T", bound=Component)


# NOTE: pydantic does not yet seem to be fully compatible with the
# new generics syntax (3.12+) so we use the old one
class ComponentConfig(BaseModel, Generic[U, T]):  # noqa: UP046
    """
    Base model defining common fields and logic for configuration and
    storage of a UI component.

    Attributes
    ----------
    label : str | None, default=None
        The label of the component.
    value : U
        The default value of the component.
    visible : bool, default=True
        Whether the component is visible.
    scale : int | None, default=None
        The scale of the component.
    render : bool, default=True
        Whether to render the component when instantiated.
    exclude_value : bool, default=False
        If True, the default value of the component cannot be updated.
    _instance: T | None, default=None
        Internal attribute storing the component instance. Will be null
        until the component is instantiated.
    instance: T
        The component instance. Attempting to access this field before
        the component is instantiated will raise a NotInstantiatedError.

    """

    label: str | None = None
    value: U
    visible: bool = True
    scale: int | None = None
    render: bool = True
    exclude_value: bool = False
    _instance: T | None = PrivateAttr(default=None)

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @property
    def instance(self) -> T:
        """
        Property which returns the component instance.

        Raises
        ------
        ComponentNotInstatiatedError
            If the component has not been instantiated yet.

        """
        if self._instance is None:
            raise ComponentNotInstatiatedError
        return self._instance

    @model_serializer(mode="wrap")
    def serialize(self, nxt: SerializerFunctionWrapHandler) -> dict[str, Any]:
        """
        Serialize the component configuration to a dictionary using a
        custom field order.

        Parameters
        ----------
        nxt : SerializerFunctionWrapHandler
            The main serialization function to apply to the component
            configuration before applying the custom field order.

        Returns
        -------
        dict[str, Any]
            The serialized component configuration.

        """
        field_order = [
            "label",
            "info",
            "value",
            "minimum",
            "maximum",
            "step",
            "precision",
            "choices",
            "multiselect",
            "allow_custom_value",
            "type",
            "visible",
            "scale",
            "render",
            "exclude_value",
        ]
        config = nxt(self)

        ordered_mapping = {key: config[key] for key in field_order if key in config}
        remaining_mapping = {
            key: value for key, value in config.items() if key not in field_order
        }
        return ordered_mapping | remaining_mapping


type AnyComponentConfig = ComponentConfig[Any, Any]


class InfoComponentConfig(ComponentConfig[U, T], Generic[U, T]):  # noqa: UP046
    """
    Configuration settings for a component with an information text.

    Attributes
    ----------
    info : str | None, default=None
        The information text on the component.

    See Also
    --------
    ComponentConfig
        Parent model defining common configuration settings for a UI
        component.

    """

    info: str | None = None


class SliderConfig(InfoComponentConfig[float | None, gr.Slider]):
    """
    Configuration settings for a slider component.

    Attributes
    ----------
    minimum : float, default=0
        The minimum value of the slider component.
    maximum : float, default=100
        The maximum value of the slider component.
    step : float | None, default=None
        The step size of the slider component.

    See Also
    --------
    InfoComponentConfig
        Parent model defining common configuration settings for a UI
        component with an information text.

    """

    minimum: float = 0
    maximum: float = 100
    step: float | None = None

    def instantiate(
        self,
        maximum: float | None = None,
        value: float | None = None,
    ) -> None:
        """
        Instantiate the slider component.

        Parameters
        ----------
        maximum : float | None, default=None
            The maximum value of the slider component. If not provided,
            the maximum value specified in the configuration is used.
        value : float | None, default=None
            The initial value of the slider component. If not provided,
            the value specified in the configuration is used.

        """
        self._instance = gr.Slider(
            label=self.label,
            info=self.info,
            value=self.value if value is None else value,
            minimum=self.minimum,
            maximum=self.maximum if maximum is None else maximum,
            step=self.step,
            visible=self.visible,
            scale=self.scale,
            render=self.render,
            show_reset_button=False,
        )

    @classmethod
    def octave_shift(cls, label: str, info: str) -> SliderConfig:
        """
        Create a slider configuration for octave pitch shift.

        Parameters
        ----------
        label : str
            The label for the octave pitch shift slider.
        info : str
            The information text for the octave pitch shift slider.

        Returns
        -------
        SliderConfig
            A slider configuration for octave pitch shift.

        """
        return cls(label=label, info=info, value=0, minimum=-3, maximum=3, step=1)

    @classmethod
    def semitone_shift(cls, label: str, info: str) -> SliderConfig:
        """
        Create a slider configuration for semitone pitch shift.

        Parameters
        ----------
        label : str
            The label for the semitone pitch shift slider.
        info : str
            The information text for the semitone pitch shift slider.

        Returns
        -------
        SliderConfig
            A slider configuration for semitone pitch shift.

        """
        return cls(label=label, info=info, value=0, minimum=-12, maximum=12, step=1)

    @classmethod
    def hop_length(cls, label: str, info: str, visible: bool) -> SliderConfig:
        """
        Create a slider configuration for hop length.

        Parameters
        ----------
        label : str
            The label for the hop length slider.
        info : str
            The information text for the hop length slider.
        visible : bool
            Whether the slider should be visible.

        Returns
        -------
        SliderConfig
            A slider configuration for hop length.

        """
        return cls(
            label=label,
            info=info,
            value=128,
            minimum=1,
            maximum=512,
            step=1,
            visible=visible,
        )

    @classmethod
    def clean_strength(cls, visible: bool) -> SliderConfig:
        """
        Create a slider configuration for clean strength.

        Parameters
        ----------
        visible : bool
            Whether the slider should be visible.

        Returns
        -------
        SliderConfig
            A slider configuration for clean strength.

        """
        return cls(
            label="Cleaning intensity",
            info=(
                "Higher values result in stronger cleaning, but may lead to a more"
                " compressed sound."
            ),
            value=0.7,
            minimum=0.0,
            maximum=1.0,
            step=0.1,
            visible=visible,
        )

    @classmethod
    def gain(cls, label: str, info: str) -> SliderConfig:
        """
        Create a slider configuration for gain.

        Parameters
        ----------
        label : str
            The label for the gain slider.
        info : str
            The information text for the gain slider.

        Returns
        -------
        SliderConfig
            A slider configuration for gain.

        """
        return cls(label=label, info=info, value=0, minimum=-20, maximum=20, step=1)

    @classmethod
    def cpu_cores(cls) -> SliderConfig:
        """
        Create a slider configuration for CPU cores.

        Returns
        -------
        SliderConfig
            A slider configuration for CPU cores.

        """
        return cls(
            label="CPU cores",
            info="The number of CPU cores to use for multi-threading.",
            value=None,
            minimum=1,
            maximum=1,
            step=1,
            exclude_value=True,
        )


class CheckboxConfig(InfoComponentConfig[bool, gr.Checkbox]):
    """
    Configuration settings for a checkbox component.

    See Also
    --------
    InfoComponentConfig
        Parent model defining common configuration settings for a UI
        component with an information text.

    """

    def instantiate(self) -> None:
        """Instantiate the checkbox component."""
        self._instance = gr.Checkbox(
            label=self.label,
            info=self.info,
            value=self.value,
            visible=self.visible,
            scale=self.scale,
            render=self.render,
        )


class NumberConfig(InfoComponentConfig[int | None, gr.Number]):
    """
    Configuration settings for a number component.

    Attributes
    ----------
    precision : int | None, default=None
        The number of decimal places to display for the number
        component.

    See Also
    --------
    InfoComponentConfig
        Parent model defining common configuration settings for a UI
        component with an information text.

    """

    precision: int | None = None

    def instantiate(self) -> None:
        """Instantiate the number component."""
        self._instance = gr.Number(
            label=self.label,
            info=self.info,
            value=self.value,
            precision=self.precision,
            visible=self.visible,
            scale=self.scale,
            render=self.render,
        )


class RadioConfig(InfoComponentConfig[BaseRadioValue, gr.Radio]):
    """
    Configuration settings for a radio component.

    Attributes
    ----------
    choices : BaseRadioValue, default=None
        The selectable choices for the radio component.

    See Also
    --------
    InfoComponentConfig
        Parent model defining common configuration settings for a UI
        component with an information text.

    """

    choices: RadioChoices = None

    def instantiate(self) -> None:
        """Instantiate the radio component."""
        self._instance = gr.Radio(
            label=self.label,
            info=self.info,
            value=self.value,
            choices=self.choices,
            visible=self.visible,
            scale=self.scale,
            render=self.render,
        )


class DropdownConfig(InfoComponentConfig[BaseDropdownValue, gr.Dropdown]):
    """
    Configuration settings for a dropdown component.

    Attributes
    ----------
    choices : DropdownChoices, default=None
        The selectable choices for the dropdown component.
    multiselect : bool | None, default=None
        Whether the dropdown component allows multiple selections.

    See Also
    --------
    ComponentConfig
        Parent model defining common configuration settings for a UI
        component.

    """

    choices: DropdownChoices = None
    multiselect: bool | None = None
    allow_custom_value: bool = False
    type: Literal["value", "index"] = "value"

    def instantiate(
        self,
        value: BaseDropdownValue = None,
        choices: DropdownChoices = None,
    ) -> None:
        """
        Instantiate the dropdown component.

        Parameters
        ----------
        value : BaseDropdownValue, default=None
            The initial value of the dropdown component. If not
            provided, the value specified in the configuration is used.
        choices : DropdownChoices, default=None
            Custom choices to instantiate the dropdown component with.
            If not provided, the dropdown will be instantiated with the
            choices specified in the configuration.

        """
        self._instance = gr.Dropdown(
            label=self.label,
            info=self.info,
            value=self.value if value is None else value,
            choices=self.choices if choices is None else choices,
            multiselect=self.multiselect,
            allow_custom_value=self.allow_custom_value,
            type=self.type,
            visible=self.visible,
            scale=self.scale,
            render=self.render,
        )

    @classmethod
    def hardware_acceleration(cls) -> DropdownConfig:
        """
        Create a dropdown configuration for hardware acceleration.

        Returns
        -------
        DropdownConfig
            A dropdown configuration for hardware acceleration.

        """
        return cls(
            label="Hardware acceleration",
            info=(
                "The type of hardware acceleration to use. 'Automatic' will"
                " automatically select the first available GPU and fall back to CPU if"
                " no GPUs are available."
            ),
            value=DeviceType.AUTOMATIC,
            choices=list(DeviceType),
            exclude_value=True,
        )

    @classmethod
    def song_dir(cls) -> DropdownConfig:
        """
        Create a song directory dropdown configuration.

        Returns
        -------
        DropdownConfig
            A song directory dropdown configuration.

        """
        return cls(
            label="Song directory",
            info=(
                "Directory where intermediate audio files are stored and loaded from"
                " locally. When a new song is retrieved, its directory is chosen by"
                " default."
            ),
            value=None,
            render=False,
            exclude_value=True,
        )

    @classmethod
    def gpu(cls) -> DropdownConfig:
        """
        Create a GPU dropdown configuration.

        Returns
        -------
        DropdownConfig
            A GPU dropdown configuration.

        """
        return cls(
            label="GPU(s)",
            info="The GPU(s) to use for hardware acceleration.",
            value=None,
            multiselect=True,
            visible=False,
            exclude_value=True,
        )

    @classmethod
    def multi_delete(cls, label: str, info: str) -> DropdownConfig:
        """
        Create a multi-delete dropdown configuration.

        Parameters
        ----------
        label : str
            The label for the multi-delete dropdown.
        info : str
            The information text for the multi-delete dropdown.

        Returns
        -------
        DropdownConfig
            A multi-delete dropdown configuration.

        """
        return cls(
            label=label,
            info=info,
            value=None,
            multiselect=True,
            render=False,
            exclude_value=True,
        )


class TextboxConfig(InfoComponentConfig[str | None, gr.Textbox]):
    """
    Configuration settings for a textbox component.

    Attributes
    ----------
    placeholder : str | None, default=None
        The placeholder text for the textbox component.

    See Also
    --------
    InfoComponentConfig
        Parent model defining common configuration settings for a UI
        component with an information text.

    """

    placeholder: str | None = None

    def instantiate(
        self,
        value: AnyCallable | None = None,
        inputs: Component | Sequence[Component] | set[Component] | None = None,
    ) -> None:
        """
        Instantiate the textbox component.

        Parameters
        ----------
        value : AnyCallable | None, default=None
            A custom function to instantiate the component's value with.
            Meant to be used in conjunction with the `inputs` parameter.
            If not provided, the textbox will be instantiated with the
            default value specified in the configuration.
        inputs : Component | Sequence[Component] | set[Component] | None
            A sequence of components whose values will be passed to the
            function specified in the `value` parameter. If not provided
            , has no effect.

        """
        self._instance = gr.Textbox(
            label=self.label,
            info=self.info,
            value=self.value if value is None else value,
            placeholder=self.placeholder,
            visible=self.visible,
            scale=self.scale,
            render=self.render,
            inputs=inputs,
        )


class AudioConfig(ComponentConfig[str | Path | None, gr.Audio]):
    """
    Configuration settings for an audio component.

    Attributes
    ----------
    interactive : bool | None, default=None
        Whether the audio component is interactive.

    See Also
    --------
    ComponentConfig
        Parent model defining common configuration settings for a UI
        component.

    """

    interactive: bool | None = None

    def instantiate(self) -> None:
        """Instantiate the audio component."""
        self._instance = gr.Audio(
            label=self.label,
            value=self.value,
            interactive=self.interactive,
            type="filepath",
            visible=self.visible,
            render=self.render,
            scale=self.scale,
            waveform_options=gr.WaveformOptions(show_recording_waveform=False),
        )

    @classmethod
    def intermediate(cls, label: str) -> AudioConfig:
        """
        Create an intermediate audio configuration.

        Returns
        -------
        AudioConfig
            An intermediate audio configuration.

        """
        return cls(label=label, value=None, exclude_value=True)

    @classmethod
    def input(cls, label: str) -> AudioConfig:
        """
        Create an input audio configuration.

        Returns
        -------
        AudioConfig
            An input audio configuration.

        """
        return cls(label=label, value=None, render=False, exclude_value=True)



import torch
import torch.utils.data

from librosa.filters import mel as librosa_mel_fn


def dynamic_range_compression_torch(x, C=1, clip_val=1e-5):
    """
    Dynamic range compression using log10.

    Args:
        x (torch.Tensor): Input tensor.
        C (float, optional): Scaling factor. Defaults to 1.
        clip_val (float, optional): Minimum value for clamping. Defaults to 1e-5.

    """
    return torch.log(torch.clamp(x, min=clip_val) * C)


def dynamic_range_decompression_torch(x, C=1):
    """
    Dynamic range decompression using exp.

    Args:
        x (torch.Tensor): Input tensor.
        C (float, optional): Scaling factor. Defaults to 1.

    """
    return torch.exp(x) / C


def spectral_normalize_torch(magnitudes):
    """
    Spectral normalization using dynamic range compression.

    Args:
        magnitudes (torch.Tensor): Magnitude spectrogram.

    """
    return dynamic_range_compression_torch(magnitudes)


def spectral_de_normalize_torch(magnitudes):
    """
    Spectral de-normalization using dynamic range decompression.

    Args:
        magnitudes (torch.Tensor): Normalized spectrogram.

    """
    return dynamic_range_decompression_torch(magnitudes)


mel_basis = {}
hann_window = {}


def spectrogram_torch(y, n_fft, hop_size, win_size, center=False):
    """
    Compute the spectrogram of a signal using STFT.

    Args:
        y (torch.Tensor): Input signal.
        n_fft (int): FFT window size.
        hop_size (int): Hop size between frames.
        win_size (int): Window size.
        center (bool, optional): Whether to center the window. Defaults to False.

    """
    global hann_window
    dtype_device = str(y.dtype) + "_" + str(y.device)
    wnsize_dtype_device = str(win_size) + "_" + dtype_device
    if wnsize_dtype_device not in hann_window:
        hann_window[wnsize_dtype_device] = torch.hann_window(win_size).to(
            dtype=y.dtype,
            device=y.device,
        )

    y = torch.nn.functional.pad(
        y.unsqueeze(1),
        (int((n_fft - hop_size) / 2), int((n_fft - hop_size) / 2)),
        mode="reflect",
    )
    y = y.squeeze(1)

    spec = torch.stft(
        y,
        n_fft=n_fft,
        hop_length=hop_size,
        win_length=win_size,
        window=hann_window[wnsize_dtype_device],
        center=center,
        pad_mode="reflect",
        normalized=False,
        onesided=True,
        return_complex=True,
    )

    spec = torch.sqrt(spec.real.pow(2) + spec.imag.pow(2) + 1e-6)

    return spec


def spec_to_mel_torch(spec, n_fft, num_mels, sample_rate, fmin, fmax):
    """
    Convert a spectrogram to a mel-spectrogram.

    Args:
        spec (torch.Tensor): Magnitude spectrogram.
        n_fft (int): FFT window size.
        num_mels (int): Number of mel frequency bins.
        sample_rate (int): Sampling rate of the audio signal.
        fmin (float): Minimum frequency.
        fmax (float): Maximum frequency.

    """
    global mel_basis
    dtype_device = str(spec.dtype) + "_" + str(spec.device)
    fmax_dtype_device = str(fmax) + "_" + dtype_device
    if fmax_dtype_device not in mel_basis:
        mel = librosa_mel_fn(
            sr=sample_rate,
            n_fft=n_fft,
            n_mels=num_mels,
            fmin=fmin,
            fmax=fmax,
        )
        mel_basis[fmax_dtype_device] = torch.from_numpy(mel).to(
            dtype=spec.dtype,
            device=spec.device,
        )

    melspec = torch.matmul(mel_basis[fmax_dtype_device], spec)
    melspec = spectral_normalize_torch(melspec)
    return melspec


def mel_spectrogram_torch(
    y,
    n_fft,
    num_mels,
    sample_rate,
    hop_size,
    win_size,
    fmin,
    fmax,
    center=False,
):
    """
    Compute the mel-spectrogram of a signal.

    Args:
        y (torch.Tensor): Input signal.
        n_fft (int): FFT window size.
        num_mels (int): Number of mel frequency bins.
        sample_rate (int): Sampling rate of the audio signal.
        hop_size (int): Hop size between frames.
        win_size (int): Window size.
        fmin (float): Minimum frequency.
        fmax (float): Maximum frequency.
        center (bool, optional): Whether to center the window. Defaults to False.

    """
    spec = spectrogram_torch(y, n_fft, hop_size, win_size, center)

    melspec = spec_to_mel_torch(spec, n_fft, num_mels, sample_rate, fmin, fmax)

    return melspec


def compute_window_length(n_mels: int, sample_rate: int):
    f_min = 0
    f_max = sample_rate / 2
    window_length_seconds = 8 * n_mels / (f_max - f_min)
    window_length = int(window_length_seconds * sample_rate)
    return 2 ** (window_length.bit_length() - 1)


class MultiScaleMelSpectrogramLoss(torch.nn.Module):

    def __init__(
        self,
        sample_rate: int = 24000,
        n_mels: list[int] = [5, 10, 20, 40, 80, 160, 320, 480],
        loss_fn=torch.nn.L1Loss(),
    ):
        super().__init__()
        self.sample_rate = sample_rate
        self.loss_fn = loss_fn
        self.log_base = torch.log(torch.tensor(10.0))
        self.stft_params: list[tuple] = []
        self.hann_window: dict[int, torch.Tensor] = {}
        self.mel_banks: dict[int, torch.Tensor] = {}

        self.stft_params = [
            (mel, compute_window_length(mel, sample_rate), self.sample_rate // 100)
            for mel in n_mels
        ]

    def mel_spectrogram(
        self,
        wav: torch.Tensor,
        n_mels: int,
        window_length: int,
        hop_length: int,
    ):
        # IDs for caching
        dtype_device = str(wav.dtype) + "_" + str(wav.device)
        win_dtype_device = str(window_length) + "_" + dtype_device
        mel_dtype_device = str(n_mels) + "_" + dtype_device
        # caching hann window
        if win_dtype_device not in self.hann_window:
            self.hann_window[win_dtype_device] = torch.hann_window(
                window_length,
                device=wav.device,
                dtype=torch.float32,
            )

        wav = wav.squeeze(1)  # -> torch(B, T)

        stft = torch.stft(
            wav.float(),
            n_fft=window_length,
            hop_length=hop_length,
            window=self.hann_window[win_dtype_device],
            return_complex=True,
        )  # -> torch (B, window_length // 2 + 1, (T - window_length)/hop_length + 1)

        magnitude = torch.sqrt(stft.real.pow(2) + stft.imag.pow(2) + 1e-6)

        # caching mel filter
        if mel_dtype_device not in self.mel_banks:
            self.mel_banks[mel_dtype_device] = torch.from_numpy(
                librosa_mel_fn(
                    sr=self.sample_rate,
                    n_mels=n_mels,
                    n_fft=window_length,
                    fmin=0,
                    fmax=None,
                ),
            ).to(device=wav.device, dtype=torch.float32)

        mel_spectrogram = torch.matmul(
            self.mel_banks[mel_dtype_device],
            magnitude,
        )  # torch(B, n_mels, stft.frames)
        return mel_spectrogram

    def forward(
        self,
        real: torch.Tensor,
        fake: torch.Tensor,
    ):  # real: torch(B, 1, T) , fake: torch(B, 1, T)
        loss = 0.0
        for p in self.stft_params:
            real_mels = self.mel_spectrogram(real, *p)
            fake_mels = self.mel_spectrogram(fake, *p)
            real_logmels = torch.log(real_mels.clamp(min=1e-5)) / self.log_base
            fake_logmels = torch.log(fake_mels.clamp(min=1e-5)) / self.log_base
            loss += self.loss_fn(real_logmels, fake_logmels)
        return loss



"""Common variables used in the Ultimate RVC project."""

from __future__ import annotations

import os
import sys
from pathlib import Path

BASE_DIR = Path.cwd()
VENV_DIR = Path(sys.prefix)
MODELS_DIR = Path(os.getenv("URVC_MODELS_DIR") or BASE_DIR / "models")
RVC_MODELS_DIR = MODELS_DIR / "rvc"
VOICE_MODELS_DIR = Path(
    os.getenv("URVC_VOICE_MODELS_DIR") or RVC_MODELS_DIR / "voice_models",
)
EMBEDDER_MODELS_DIR = RVC_MODELS_DIR / "embedders"
CUSTOM_EMBEDDER_MODELS_DIR = EMBEDDER_MODELS_DIR / "custom"
PRETRAINED_MODELS_DIR = RVC_MODELS_DIR / "pretraineds"
CUSTOM_PRETRAINED_MODELS_DIR = PRETRAINED_MODELS_DIR / "custom"

SEPARATOR_MODELS_DIR = MODELS_DIR / "audio_separator"
TRAINING_MODELS_DIR = RVC_MODELS_DIR / "training"
AUDIO_DIR = Path(os.getenv("URVC_AUDIO_DIR") or BASE_DIR / "audio")
TEMP_DIR = Path(os.getenv("URVC_TEMP_DIR") or BASE_DIR / "temp")
CONFIG_DIR = Path(os.getenv("URVC_CONFIG_DIR") or BASE_DIR / "config")
NODE_PATH = Path(
    (
        os.getenv("GRADIO_NODE_PATH")
        or (
            VENV_DIR
            / f"lib/python{sys.version_info.major}.{sys.version_info.minor}"
            / "site-packages/nodejs_wheel/bin/node"
        )
        if sys.platform == "linux"
        else VENV_DIR / "Lib/site-packages/nodejs_wheel/node.exe"
    ),
)



import json
import logging
import os
from multiprocessing import cpu_count

from sklearn.cluster import MiniBatchKMeans

import numpy as np

import faiss

logger = logging.getLogger(__name__)


def main(exp_dir: str, index_algorithm: str) -> None:

    try:
        model_info = json.load(open(os.path.join(exp_dir, "model_info.json")))
        embedder_model = model_info["embedder_model"]
        custom_embedder_model_hash = model_info.get("custom_embedder_model_hash", None)
        if custom_embedder_model_hash is not None:
            embedder_model = f"custom_{custom_embedder_model_hash}"
        feature_dir = os.path.join(exp_dir, f"{embedder_model}_extracted")
        model_name = os.path.basename(exp_dir)

        index_filename_added = f"{model_name}.index"
        index_filepath_added = os.path.join(exp_dir, index_filename_added)

        if os.path.exists(index_filepath_added):
            pass
        else:
            npys = []
            listdir_res = sorted(os.listdir(feature_dir))

            for name in listdir_res:
                file_path = os.path.join(feature_dir, name)
                phone = np.load(file_path)
                npys.append(phone)

            big_npy = np.concatenate(npys, axis=0)

            big_npy_idx = np.arange(big_npy.shape[0])
            np.random.shuffle(big_npy_idx)
            big_npy = big_npy[big_npy_idx]

            if big_npy.shape[0] > 2e5 and (
                index_algorithm == "Auto" or index_algorithm == "KMeans"
            ):
                big_npy = (
                    MiniBatchKMeans(
                        n_clusters=10000,
                        verbose=True,
                        batch_size=256 * cpu_count(),
                        compute_labels=False,
                        init="random",
                    )
                    .fit(big_npy)
                    .cluster_centers_
                )

            n_ivf = min(int(16 * np.sqrt(big_npy.shape[0])), big_npy.shape[0] // 39)

            # index_added
            index_added = faiss.index_factory(768, f"IVF{n_ivf},Flat")
            index_ivf_added = faiss.extract_index_ivf(index_added)
            index_ivf_added.nprobe = 1
            index_added.train(big_npy)

            batch_size_add = 8192
            for i in range(0, big_npy.shape[0], batch_size_add):
                index_added.add(big_npy[i : i + batch_size_add])

            faiss.write_index(index_added, index_filepath_added)
            logger.info("Saved index file '%s'", index_filepath_added)

    except Exception as error:
        logger.error(  # noqa: TRY400
            "An error occurred extracting the index: %s. If you are running this code"
            " in a virtual environment, make sure you have enough GPU available to"
            " generate the Index file.",
            error,
        )



"""
Module which defines the command-line interface for using RVC-based
TTS.
"""

from __future__ import annotations

from typing import Annotated

import time

import typer
from rich import print as rprint
from rich.panel import Panel
from rich.table import Table

from ultimate_rvc.cli.common import (
    complete_audio_ext,
    complete_embedder_model,
    complete_f0_method,
    format_duration,
)
from ultimate_rvc.cli.typing_extra import PanelName
from ultimate_rvc.core.generate import speech as generate_speech
from ultimate_rvc.typing_extra import AudioExt, EmbedderModel, F0Method

app = typer.Typer(
    name="speech",
    no_args_is_help=True,
    help="Generate speech from text using RVC.",
    rich_markup_mode="markdown",
)


@app.command(no_args_is_help=True)
def run_edge_tts(
    source: Annotated[
        str,
        typer.Argument(
            help="A string or path to a file containing the text to be converted.",
        ),
    ],
    voice: Annotated[
        str,
        typer.Option(
            help=(
                "The short name of the Edge TTS voice which should speak the provided"
                " text. Use the `list-edge-voices` command to get a list of available"
                " Edge TTS voices."
            ),
        ),
    ] = "en-US-ChristopherNeural",
    pitch_shift: Annotated[
        int,
        typer.Option(
            help=(
                "The number of hertz to shift the pitch of the Edge TTS voice"
                " speaking the provided text."
            ),
        ),
    ] = 0,
    speed_change: Annotated[
        int,
        typer.Option(
            help=(
                "The percentual change to the speed of the Edge TTS voice speaking the"
                " provided text."
            ),
        ),
    ] = 0,
    volume_change: Annotated[
        int,
        typer.Option(
            help=(
                "The percentual change to the volume of the Efge TTS voice speaking the"
                " provided text."
            ),
        ),
    ] = 0,
) -> None:
    """Convert text to speech using Edge TTS."""
    start_time = time.perf_counter()

    rprint()

    audio_path = generate_speech.run_edge_tts(
        source,
        voice,
        pitch_shift,
        speed_change,
        volume_change,
    )

    rprint("[+] Text successfully converted to speech!")
    rprint()
    rprint("Elapsed time:", format_duration(time.perf_counter() - start_time))
    rprint(Panel(f"[green]{audio_path}", title="Speech Path"))


@app.command()
def list_edge_tts_voices(
    locale: Annotated[
        str | None,
        typer.Option(
            help="The locale to filter Edge TTS voices by.",
        ),
    ] = None,
    content_category: Annotated[
        list[str] | None,
        typer.Option(
            help=(
                "The content category to filter Edge TTS voices by. This option can be"
                " supplied multiple times to filter by multiple content categories."
            ),
        ),
    ] = None,
    voice_personality: Annotated[
        list[str] | None,
        typer.Option(
            help=(
                "The voice personality to filter Edge TTS voices by. This option can be"
                " supplied multiple times to filter by multiple voice personalities."
            ),
        ),
    ] = None,
    offset: Annotated[
        int,
        typer.Option(
            min=0,
            help="The offset to start listing Edge TTS voices from.",
        ),
    ] = 0,
    limit: Annotated[
        int,
        typer.Option(
            min=0,
            help="The limit on how many Edge TTS voices to list.",
        ),
    ] = 20,
    include_status_info: Annotated[
        bool,
        typer.Option(
            help="Include status information for each Edge TTS voice.",
        ),
    ] = False,
) -> None:
    """List Edge TTS voices based on provided filters."""
    start_time = time.perf_counter()

    rprint()
    rprint("[~] Retrieving information on all available edge TTS voices...")

    voices, keys = generate_speech.list_edge_tts_voices(
        locale=locale,
        content_categories=content_category,
        voice_personalities=voice_personality,
        offset=offset,
        limit=limit,
        include_status_info=include_status_info,
    )

    table = Table()
    for key in keys:
        table.add_column(key)
    for voice in voices:
        table.add_row(*[f"[green]{voice_attrib}" for voice_attrib in voice])

    rprint("[+] Information successfully retrieved!")
    rprint()

    rprint("Elapsed time:", format_duration(time.perf_counter() - start_time))
    rprint(Panel(table, title="Available Edge TTS Voices"))


@app.command(no_args_is_help=True)
def run_pipeline(
    source: Annotated[
        str,
        typer.Argument(
            help="A string or path to a file containing the text to be converted.",
        ),
    ],
    model_name: Annotated[
        str,
        typer.Argument(
            help="The name of the RVC model to use for speech conversion.",
        ),
    ],
    tts_voice: Annotated[
        str,
        typer.Option(
            help="The short name of the Edge TTS voice to use for text-to-speech"
            " conversion. Use the `list-edge-voices` command to get a list of available"
            " Edge TTS voices.",
            rich_help_panel=PanelName.EDGE_TTS_OPTIONS,
        ),
    ] = "en-US-ChristopherNeural",
    tts_pitch_shift: Annotated[
        int,
        typer.Option(
            help=(
                "The number of hertz to shift the pitch of the speech generated"
                " by Edge TTS."
            ),
            rich_help_panel=PanelName.EDGE_TTS_OPTIONS,
        ),
    ] = 0,
    tts_speed_change: Annotated[
        int,
        typer.Option(
            help=(
                "The percentual change to the speed of the speech generated by Edge"
                " TTS."
            ),
            rich_help_panel=PanelName.EDGE_TTS_OPTIONS,
        ),
    ] = 0,
    tts_volume_change: Annotated[
        int,
        typer.Option(
            help=(
                "The percentual change to the volume of the speech generated by Edge"
                " TTS."
            ),
            rich_help_panel=PanelName.EDGE_TTS_OPTIONS,
        ),
    ] = 0,
    n_octaves: Annotated[
        int,
        typer.Option(
            help=(
                "The number of octaves to shift the pitch of the speech converted using"
                " RVC. Use 1 for male-to-female and -1 for vice-versa."
            ),
            rich_help_panel=PanelName.RVC_MAIN_OPTIONS,
        ),
    ] = 0,
    n_semitones: Annotated[
        int,
        typer.Option(
            help=(
                "The number of semitones to shift the pitch of the speech converted"
                " using RVC. Altering this slightly reduces sound quality."
            ),
            rich_help_panel=PanelName.RVC_MAIN_OPTIONS,
        ),
    ] = 0,
    f0_method: Annotated[
        list[F0Method] | None,
        typer.Option(
            case_sensitive=False,
            autocompletion=complete_f0_method,
            rich_help_panel=PanelName.RVC_SYNTHESIS_OPTIONS,
            help=(
                "The method to use for pitch extraction during the RVC process. This"
                " option can be provided multiple times to use multiple pitch"
                " extraction methods in combination. If not provided, will default to"
                " the rmvpe method, which is generally recommended."
            ),
        ),
    ] = None,
    index_rate: Annotated[
        float,
        typer.Option(
            min=0,
            max=1,
            rich_help_panel=PanelName.RVC_SYNTHESIS_OPTIONS,
            help=(
                "The rate of influence of the RVC index file. Increase to"
                " bias the conversion towards the accent of the used voice model."
                " Decrease to potentially reduce artifacts."
            ),
        ),
    ] = 0.3,
    rms_mix_rate: Annotated[
        float,
        typer.Option(
            min=0,
            max=1,
            rich_help_panel=PanelName.RVC_SYNTHESIS_OPTIONS,
            help=(
                "Blending rate for the volume envelope of the speech converted using"
                " RVC. Controls how much to mimic the loudness of the given Edge TTS"
                " speech (0) or a fixed loudness (1)."
            ),
        ),
    ] = 1.0,
    protect_rate: Annotated[
        float,
        typer.Option(
            min=0,
            max=0.5,
            rich_help_panel=PanelName.RVC_SYNTHESIS_OPTIONS,
            help=(
                "A coefficient which controls the extent to which consonants and"
                " breathing sounds are protected from artifacts during the RVC"
                " process. A higher value offers more protection but may worsen the"
                " indexing effect."
            ),
        ),
    ] = 0.33,
    hop_length: Annotated[
        int,
        typer.Option(
            min=1,
            max=512,
            rich_help_panel=PanelName.RVC_SYNTHESIS_OPTIONS,
            help=(
                "Controls how often the CREPE-based pitch extraction method checks for"
                " pitch changes during the RVC process. Measured in milliseconds."
                " Lower values lead to longer conversion times and a higher risk of"
                " voice cracks, but better pitch accuracy."
            ),
        ),
    ] = 128,
    split_speech: Annotated[
        bool,
        typer.Option(
            rich_help_panel=PanelName.RVC_ENRICHMENT_OPTIONS,
            help=(
                "Whether to split the Edge TTS speech into smaller segments before"
                " converting it using RVC. This can improve output quality for longer"
                " speech."
            ),
        ),
    ] = True,
    autotune_speech: Annotated[
        bool,
        typer.Option(
            rich_help_panel=PanelName.RVC_ENRICHMENT_OPTIONS,
            help="Whether to apply autotune to the speech converted using RVC.",
        ),
    ] = False,
    autotune_strength: Annotated[
        float,
        typer.Option(
            min=0,
            max=1,
            rich_help_panel=PanelName.RVC_ENRICHMENT_OPTIONS,
            help=(
                "The intensity of the autotune effect to apply to the speech converted"
                " using RVC. Higher values result in stronger snapping to the chromatic"
                " grid."
            ),
        ),
    ] = 1.0,
    clean_speech: Annotated[
        bool,
        typer.Option(
            rich_help_panel=PanelName.RVC_ENRICHMENT_OPTIONS,
            help=(
                "Whether to clean the speech converted using RVC using noise reduction"
                " algorithms"
            ),
        ),
    ] = True,
    clean_strength: Annotated[
        float,
        typer.Option(
            min=0,
            max=1,
            rich_help_panel=PanelName.RVC_ENRICHMENT_OPTIONS,
            help=(
                "The intensity of the cleaning to apply to the speech converted using"
                " RVC. Higher values result in stronger cleaning, but may lead to a"
                " more compressed sound."
            ),
        ),
    ] = 0.7,
    embedder_model: Annotated[
        EmbedderModel,
        typer.Option(
            case_sensitive=False,
            autocompletion=complete_embedder_model,
            rich_help_panel=PanelName.RVC_EMBEDDINGS_OPTIONS,
            help=(
                "The model to use for generating speaker embeddings during the RVC"
                " process."
            ),
        ),
    ] = EmbedderModel.CONTENTVEC,
    custom_embedder_model: Annotated[
        str | None,
        typer.Option(
            rich_help_panel=PanelName.RVC_EMBEDDINGS_OPTIONS,
            help=(
                "The name of a custom embedder model to use for generating speaker"
                " embeddings during the RVC process. Only applicable if"
                " `embedder-model` is set to `custom`."
            ),
        ),
    ] = None,
    sid: Annotated[
        int,
        typer.Option(
            rich_help_panel=PanelName.RVC_EMBEDDINGS_OPTIONS,
            help="The id of the speaker to use for multi-speaker RVC models.",
        ),
    ] = 0,
    output_gain: Annotated[
        int,
        typer.Option(
            rich_help_panel=PanelName.AUDIO_MIXING_OPTIONS,
            help="The gain to apply to the converted speech during mixing.",
        ),
    ] = 0,
    output_sr: Annotated[
        int,
        typer.Option(
            rich_help_panel=PanelName.AUDIO_MIXING_OPTIONS,
            help="The sample rate of the mixed speech track.",
        ),
    ] = 44100,
    output_format: Annotated[
        AudioExt,
        typer.Option(
            case_sensitive=False,
            rich_help_panel=PanelName.AUDIO_MIXING_OPTIONS,
            autocompletion=complete_audio_ext,
            help="The format of the mixed speech track.",
        ),
    ] = AudioExt.MP3,
    output_name: Annotated[
        str | None,
        typer.Option(
            rich_help_panel=PanelName.AUDIO_MIXING_OPTIONS,
            help="The name of the mixed speech track.",
        ),
    ] = None,
) -> None:
    """
    Convert text to speech using a cascaded pipeline combining Edge TTS
    and RVC. The text is first converted to speech using Edge TTS, and
    then that speech is converted to a different voice using RVC.
    """
    start_time = time.perf_counter()

    rprint()

    [output_audio_path, *intermediate_audio_file_paths] = generate_speech.run_pipeline(
        source=source,
        model_name=model_name,
        tts_voice=tts_voice,
        tts_pitch_shift=tts_pitch_shift,
        tts_speed_change=tts_speed_change,
        tts_volume_change=tts_volume_change,
        n_octaves=n_octaves,
        n_semitones=n_semitones,
        f0_methods=f0_method,
        index_rate=index_rate,
        rms_mix_rate=rms_mix_rate,
        protect_rate=protect_rate,
        hop_length=hop_length,
        split_speech=split_speech,
        autotune_speech=autotune_speech,
        autotune_strength=autotune_strength,
        clean_speech=clean_speech,
        clean_strength=clean_strength,
        embedder_model=embedder_model,
        custom_embedder_model=custom_embedder_model,
        sid=sid,
        output_gain=output_gain,
        output_sr=output_sr,
        output_format=output_format,
        output_name=output_name,
    )

    table = Table()
    table.add_column("Type")
    table.add_column("Path")
    for name, path in zip(
        [
            "Speech",
            "Converted Speech",
        ],
        intermediate_audio_file_paths,
        strict=True,
    ):
        table.add_row(name, f"[green]{path}")

    rprint("[+] Text successfully converted to speech!")
    rprint()
    rprint("Elapsed time:", format_duration(time.perf_counter() - start_time))
    rprint(Panel(f"[green]{output_audio_path}", title="Mixed Speech Path"))
    rprint(Panel(table, title="Intermediate Audio Files"))



from typing import Optional

import logging

import torch

from ultimate_rvc.rvc.lib.algorithm.commons import rand_slice_segments, slice_segments
from ultimate_rvc.rvc.lib.algorithm.encoders import PosteriorEncoder, TextEncoder
from ultimate_rvc.rvc.lib.algorithm.generators.hifigan import HiFiGANGenerator
from ultimate_rvc.rvc.lib.algorithm.generators.hifigan_mrf import HiFiGANMRFGenerator
from ultimate_rvc.rvc.lib.algorithm.generators.hifigan_nsf import HiFiGANNSFGenerator
from ultimate_rvc.rvc.lib.algorithm.generators.refinegan import RefineGANGenerator
from ultimate_rvc.rvc.lib.algorithm.residuals import ResidualCouplingBlock

logger = logging.getLogger(__name__)


class Synthesizer(torch.nn.Module):
    """
    Base Synthesizer model.

    Args:
        spec_channels (int): Number of channels in the spectrogram.
        segment_size (int): Size of the audio segment.
        inter_channels (int): Number of channels in the intermediate layers.
        hidden_channels (int): Number of channels in the hidden layers.
        filter_channels (int): Number of channels in the filter layers.
        n_heads (int): Number of attention heads.
        n_layers (int): Number of layers in the encoder.
        kernel_size (int): Size of the convolution kernel.
        p_dropout (float): Dropout probability.
        resblock (str): Type of residual block.
        resblock_kernel_sizes (list): Kernel sizes for the residual blocks.
        resblock_dilation_sizes (list): Dilation sizes for the residual blocks.
        upsample_rates (list): Upsampling rates for the decoder.
        upsample_initial_channel (int): Number of channels in the initial upsampling layer.
        upsample_kernel_sizes (list): Kernel sizes for the upsampling layers.
        spk_embed_dim (int): Dimension of the speaker embedding.
        gin_channels (int): Number of channels in the global conditioning vector.
        sr (int): Sampling rate of the audio.
        use_f0 (bool): Whether to use F0 information.
        text_enc_hidden_dim (int): Hidden dimension for the text encoder.
        kwargs: Additional keyword arguments.

    """

    def __init__(
        self,
        spec_channels: int,
        segment_size: int,
        inter_channels: int,
        hidden_channels: int,
        filter_channels: int,
        n_heads: int,
        n_layers: int,
        kernel_size: int,
        p_dropout: float,
        resblock: str,
        resblock_kernel_sizes: list,
        resblock_dilation_sizes: list,
        upsample_rates: list,
        upsample_initial_channel: int,
        upsample_kernel_sizes: list,
        spk_embed_dim: int,
        gin_channels: int,
        sr: int,
        use_f0: bool,
        text_enc_hidden_dim: int = 768,
        vocoder: str = "HiFi-GAN",
        randomized: bool = True,
        checkpointing: bool = False,
        **kwargs,
    ):
        super().__init__()
        self.segment_size = segment_size
        self.use_f0 = use_f0
        self.randomized = randomized

        self.enc_p = TextEncoder(
            inter_channels,
            hidden_channels,
            filter_channels,
            n_heads,
            n_layers,
            kernel_size,
            p_dropout,
            text_enc_hidden_dim,
            f0=use_f0,
        )
        logger.info("Using %s vocoder", vocoder)
        if use_f0:
            if vocoder == "MRF HiFi-GAN":
                self.dec = HiFiGANMRFGenerator(
                    in_channel=inter_channels,
                    upsample_initial_channel=upsample_initial_channel,
                    upsample_rates=upsample_rates,
                    upsample_kernel_sizes=upsample_kernel_sizes,
                    resblock_kernel_sizes=resblock_kernel_sizes,
                    resblock_dilations=resblock_dilation_sizes,
                    gin_channels=gin_channels,
                    sample_rate=sr,
                    harmonic_num=8,
                    checkpointing=checkpointing,
                )
            elif vocoder == "RefineGAN":
                self.dec = RefineGANGenerator(
                    sample_rate=sr,
                    downsample_rates=upsample_rates[::-1],
                    upsample_rates=upsample_rates,
                    start_channels=16,
                    num_mels=inter_channels,
                    checkpointing=checkpointing,
                )
            else:
                self.dec = HiFiGANNSFGenerator(
                    inter_channels,
                    resblock_kernel_sizes,
                    resblock_dilation_sizes,
                    upsample_rates,
                    upsample_initial_channel,
                    upsample_kernel_sizes,
                    gin_channels=gin_channels,
                    sr=sr,
                    checkpointing=checkpointing,
                )
        elif vocoder == "MRF HiFi-GAN":
            print("MRF HiFi-GAN does not support training without pitch guidance.")
            self.dec = None
        elif vocoder == "RefineGAN":
            print("RefineGAN does not support training without pitch guidance.")
            self.dec = None
        else:
            self.dec = HiFiGANGenerator(
                inter_channels,
                resblock_kernel_sizes,
                resblock_dilation_sizes,
                upsample_rates,
                upsample_initial_channel,
                upsample_kernel_sizes,
                gin_channels=gin_channels,
                checkpointing=checkpointing,
            )
        self.enc_q = PosteriorEncoder(
            spec_channels,
            inter_channels,
            hidden_channels,
            5,
            1,
            16,
            gin_channels=gin_channels,
        )
        self.flow = ResidualCouplingBlock(
            inter_channels,
            hidden_channels,
            5,
            1,
            3,
            gin_channels=gin_channels,
        )
        self.emb_g = torch.nn.Embedding(spk_embed_dim, gin_channels)

    def _remove_weight_norm_from(self, module):
        for hook in module._forward_pre_hooks.values():
            if getattr(hook, "__class__", None).__name__ == "WeightNorm":
                torch.nn.utils.remove_weight_norm(module)

    def remove_weight_norm(self):
        for module in [self.dec, self.flow, self.enc_q]:
            self._remove_weight_norm_from(module)

    def __prepare_scriptable__(self):
        self.remove_weight_norm()
        return self

    def forward(
        self,
        phone: torch.Tensor,
        phone_lengths: torch.Tensor,
        pitch: torch.Tensor | None = None,
        pitchf: torch.Tensor | None = None,
        y: torch.Tensor | None = None,
        y_lengths: torch.Tensor | None = None,
        ds: torch.Tensor | None = None,
    ):
        g = self.emb_g(ds).unsqueeze(-1)
        m_p, logs_p, x_mask = self.enc_p(phone, pitch, phone_lengths)

        if y is not None:
            z, m_q, logs_q, y_mask = self.enc_q(y, y_lengths, g=g)
            z_p = self.flow(z, y_mask, g=g)
            # regular old training method using random slices
            if self.randomized:
                z_slice, ids_slice = rand_slice_segments(
                    z,
                    y_lengths,
                    self.segment_size,
                )
                if self.use_f0:
                    pitchf = slice_segments(pitchf, ids_slice, self.segment_size, 2)
                    o = self.dec(z_slice, pitchf, g=g)
                else:
                    o = self.dec(z_slice, g=g)
                return o, ids_slice, x_mask, y_mask, (z, z_p, m_p, logs_p, m_q, logs_q)
            # future use for finetuning using the entire dataset each pass
            if self.use_f0:
                o = self.dec(z, pitchf, g=g)
            else:
                o = self.dec(z, g=g)
            return o, None, x_mask, y_mask, (z, z_p, m_p, logs_p, m_q, logs_q)
        return None, None, x_mask, None, (None, None, m_p, logs_p, None, None)

    @torch.jit.export
    def infer(
        self,
        phone: torch.Tensor,
        phone_lengths: torch.Tensor,
        pitch: torch.Tensor | None = None,
        nsff0: torch.Tensor | None = None,
        sid: torch.Tensor = None,
        rate: torch.Tensor | None = None,
    ):
        """
        Inference of the model.

        Args:
            phone (torch.Tensor): Phoneme sequence.
            phone_lengths (torch.Tensor): Lengths of the phoneme sequences.
            pitch (torch.Tensor, optional): Pitch sequence.
            nsff0 (torch.Tensor, optional): Fine-grained pitch sequence.
            sid (torch.Tensor): Speaker embedding.
            rate (torch.Tensor, optional): Rate for time-stretching.

        """
        g = self.emb_g(sid).unsqueeze(-1)
        m_p, logs_p, x_mask = self.enc_p(phone, pitch, phone_lengths)
        z_p = (m_p + torch.exp(logs_p) * torch.randn_like(m_p) * 0.66666) * x_mask

        if rate is not None:
            head = int(z_p.shape[2] * (1.0 - rate.item()))
            z_p, x_mask = z_p[:, :, head:], x_mask[:, :, head:]
            if self.use_f0 and nsff0 is not None:
                nsff0 = nsff0[:, head:]

        z = self.flow(z_p, x_mask, g=g, reverse=True)
        o = (
            self.dec(z * x_mask, nsff0, g=g)
            if self.use_f0
            else self.dec(z * x_mask, g=g)
        )

        return o, x_mask, (z, z_p, m_p, logs_p)



import numpy as np


class Slicer:
    """
    A class for slicing audio waveforms into segments based on silence detection.

    Attributes:
        sr (int): Sampling rate of the audio waveform.
        threshold (float): RMS threshold for silence detection, in dB.
        min_length (int): Minimum length of a segment, in milliseconds.
        min_interval (int): Minimum interval between segments, in milliseconds.
        hop_size (int): Hop size for RMS calculation, in milliseconds.
        max_sil_kept (int): Maximum length of silence to keep at the beginning or end of a segment, in milliseconds.

    Methods:
        slice(waveform): Slices the given waveform into segments.

    """

    def __init__(
        self,
        sr: int,
        threshold: float = -40.0,
        min_length: int = 5000,
        min_interval: int = 300,
        hop_size: int = 20,
        max_sil_kept: int = 5000,
    ):
        """
        Initializes a Slicer object.

        Args:
            sr (int): Sampling rate of the audio waveform.
            threshold (float, optional): RMS threshold for silence detection, in dB. Defaults to -40.0.
            min_length (int, optional): Minimum length of a segment, in milliseconds. Defaults to 5000.
            min_interval (int, optional): Minimum interval between segments, in milliseconds. Defaults to 300.
            hop_size (int, optional): Hop size for RMS calculation, in milliseconds. Defaults to 20.
            max_sil_kept (int, optional): Maximum length of silence to keep at the beginning or end of a segment, in milliseconds. Defaults to 5000.

        Raises:
            ValueError: If the input parameters are not valid.

        """
        if not min_length >= min_interval >= hop_size:
            raise ValueError("min_length >= min_interval >= hop_size is required")
        if not max_sil_kept >= hop_size:
            raise ValueError("max_sil_kept >= hop_size is required")

        # Convert time-based parameters to sample-based parameters
        min_interval = sr * min_interval / 1000
        self.threshold = 10 ** (threshold / 20.0)
        self.hop_size = round(sr * hop_size / 1000)
        self.win_size = min(round(min_interval), 4 * self.hop_size)
        self.min_length = round(sr * min_length / 1000 / self.hop_size)
        self.min_interval = round(min_interval / self.hop_size)
        self.max_sil_kept = round(sr * max_sil_kept / 1000 / self.hop_size)

    def _apply_slice(self, waveform, begin, end):
        """
        Applies a slice to the waveform.

        Args:
            waveform (numpy.ndarray): The waveform to slice.
            begin (int): Start frame index.
            end (int): End frame index.

        """
        start_idx = begin * self.hop_size
        if len(waveform.shape) > 1:
            end_idx = min(waveform.shape[1], end * self.hop_size)
            return waveform[:, start_idx:end_idx]
        end_idx = min(waveform.shape[0], end * self.hop_size)
        return waveform[start_idx:end_idx]

    def slice(self, waveform):
        """
        Slices the given waveform into segments.

        Args:
            waveform (numpy.ndarray): The waveform to slice.

        """
        # Calculate RMS for each frame
        samples = waveform.mean(axis=0) if len(waveform.shape) > 1 else waveform
        if samples.shape[0] <= self.min_length:
            return [waveform]

        rms_list = get_rms(
            y=samples,
            frame_length=self.win_size,
            hop_length=self.hop_size,
        ).squeeze(0)

        # Detect silence segments and mark them
        sil_tags = []
        silence_start, clip_start = None, 0
        for i, rms in enumerate(rms_list):
            # If current frame is silent
            if rms < self.threshold:
                if silence_start is None:
                    silence_start = i
                continue

            # If current frame is not silent
            if silence_start is None:
                continue

            # Check if current silence segment is leading silence or need to slice
            is_leading_silence = silence_start == 0 and i > self.max_sil_kept
            need_slice_middle = (
                i - silence_start >= self.min_interval
                and i - clip_start >= self.min_length
            )

            # If not leading silence and not need to slice middle
            if not is_leading_silence and not need_slice_middle:
                silence_start = None
                continue

            # Handle different cases of silence segments
            if i - silence_start <= self.max_sil_kept:
                # Short silence
                pos = rms_list[silence_start : i + 1].argmin() + silence_start
                if silence_start == 0:
                    sil_tags.append((0, pos))
                else:
                    sil_tags.append((pos, pos))
                clip_start = pos
            elif i - silence_start <= self.max_sil_kept * 2:
                # Medium silence
                pos = rms_list[
                    i - self.max_sil_kept : silence_start + self.max_sil_kept + 1
                ].argmin()
                pos += i - self.max_sil_kept
                pos_l = (
                    rms_list[
                        silence_start : silence_start + self.max_sil_kept + 1
                    ].argmin()
                    + silence_start
                )
                pos_r = (
                    rms_list[i - self.max_sil_kept : i + 1].argmin()
                    + i
                    - self.max_sil_kept
                )
                if silence_start == 0:
                    sil_tags.append((0, pos_r))
                    clip_start = pos_r
                else:
                    sil_tags.append((min(pos_l, pos), max(pos_r, pos)))
                    clip_start = max(pos_r, pos)
            else:
                # Long silence
                pos_l = (
                    rms_list[
                        silence_start : silence_start + self.max_sil_kept + 1
                    ].argmin()
                    + silence_start
                )
                pos_r = (
                    rms_list[i - self.max_sil_kept : i + 1].argmin()
                    + i
                    - self.max_sil_kept
                )
                if silence_start == 0:
                    sil_tags.append((0, pos_r))
                else:
                    sil_tags.append((pos_l, pos_r))
                clip_start = pos_r
            silence_start = None

        # Handle trailing silence
        total_frames = rms_list.shape[0]
        if (
            silence_start is not None
            and total_frames - silence_start >= self.min_interval
        ):
            silence_end = min(total_frames, silence_start + self.max_sil_kept)
            pos = rms_list[silence_start : silence_end + 1].argmin() + silence_start
            sil_tags.append((pos, total_frames + 1))

        # Extract segments based on silence tags
        if not sil_tags:
            return [waveform]
        chunks = []
        if sil_tags[0][0] > 0:
            chunks.append(self._apply_slice(waveform, 0, sil_tags[0][0]))

        for i in range(len(sil_tags) - 1):
            chunks.append(
                self._apply_slice(waveform, sil_tags[i][1], sil_tags[i + 1][0]),
            )

        if sil_tags[-1][1] < total_frames:
            chunks.append(
                self._apply_slice(waveform, sil_tags[-1][1], total_frames),
            )

        return chunks


def get_rms(
    y,
    frame_length=2048,
    hop_length=512,
    pad_mode="constant",
):
    """
    Calculates the root mean square (RMS) of a waveform.

    Args:
        y (numpy.ndarray): The waveform.
        frame_length (int, optional): The length of the frame in samples. Defaults to 2048.
        hop_length (int, optional): The hop length between frames in samples. Defaults to 512.
        pad_mode (str, optional): The padding mode used for the waveform. Defaults to "constant".

    """
    padding = (int(frame_length // 2), int(frame_length // 2))
    y = np.pad(y, padding, mode=pad_mode)

    axis = -1
    out_strides = y.strides + tuple([y.strides[axis]])
    x_shape_trimmed = list(y.shape)
    x_shape_trimmed[axis] -= frame_length - 1
    out_shape = tuple(x_shape_trimmed) + tuple([frame_length])
    xw = np.lib.stride_tricks.as_strided(y, shape=out_shape, strides=out_strides)

    if axis < 0:
        target_axis = axis - 1
    else:
        target_axis = axis + 1

    xw = np.moveaxis(xw, -1, target_axis)
    slices = [slice(None)] * xw.ndim
    slices[axis] = slice(0, None, hop_length)
    x = xw[tuple(slices)]

    power = np.mean(np.abs(x) ** 2, axis=-2, keepdims=True)
    return np.sqrt(power)



from typing import Optional

import torch


def init_weights(m, mean=0.0, std=0.01):
    """
    Initialize the weights of a module.

    Args:
        m: The module to initialize.
        mean: The mean of the normal distribution.
        std: The standard deviation of the normal distribution.

    """
    classname = m.__class__.__name__
    if classname.find("Conv") != -1:
        m.weight.data.normal_(mean, std)


def get_padding(kernel_size, dilation=1):
    """
    Calculate the padding needed for a convolution.

    Args:
        kernel_size: The size of the kernel.
        dilation: The dilation of the convolution.

    """
    return int((kernel_size * dilation - dilation) / 2)


def convert_pad_shape(pad_shape):
    """
    Convert the pad shape to a list of integers.

    Args:
        pad_shape: The pad shape..

    """
    l = pad_shape[::-1]
    pad_shape = [item for sublist in l for item in sublist]
    return pad_shape


def slice_segments(
    x: torch.Tensor,
    ids_str: torch.Tensor,
    segment_size: int = 4,
    dim: int = 2,
):
    """
    Slice segments from a tensor, handling tensors with different numbers of dimensions.

    Args:
        x (torch.Tensor): The tensor to slice.
        ids_str (torch.Tensor): The starting indices of the segments.
        segment_size (int, optional): The size of each segment. Defaults to 4.
        dim (int, optional): The dimension to slice across (2D or 3D tensors). Defaults to 2.

    """
    if dim == 2:
        ret = torch.zeros_like(x[:, :segment_size])
    elif dim == 3:
        ret = torch.zeros_like(x[:, :, :segment_size])

    for i in range(x.size(0)):
        idx_str = ids_str[i].item()
        idx_end = idx_str + segment_size
        if dim == 2:
            ret[i] = x[i, idx_str:idx_end]
        else:
            ret[i] = x[i, :, idx_str:idx_end]

    return ret


def rand_slice_segments(x, x_lengths=None, segment_size=4):
    """
    Randomly slice segments from a tensor.

    Args:
        x: The tensor to slice.
        x_lengths: The lengths of the sequences.
        segment_size: The size of each segment.

    """
    b, d, t = x.size()
    if x_lengths is None:
        x_lengths = t
    ids_str_max = x_lengths - segment_size + 1
    ids_str = (torch.rand([b], device=x.device) * ids_str_max).to(dtype=torch.long)
    ret = slice_segments(x, ids_str, segment_size, dim=3)
    return ret, ids_str


@torch.jit.script
def fused_add_tanh_sigmoid_multiply(input_a, input_b, n_channels):
    """
    Fused add tanh sigmoid multiply operation.

    Args:
        input_a: The first input tensor.
        input_b: The second input tensor.
        n_channels: The number of channels.

    """
    n_channels_int = n_channels[0]
    in_act = input_a + input_b
    t_act = torch.tanh(in_act[:, :n_channels_int, :])
    s_act = torch.sigmoid(in_act[:, n_channels_int:, :])
    acts = t_act * s_act
    return acts


def sequence_mask(length: torch.Tensor, max_length: int | None = None):
    """
    Generate a sequence mask.

    Args:
        length: The lengths of the sequences.
        max_length: The maximum length of the sequences.

    """
    if max_length is None:
        max_length = length.max()
    x = torch.arange(max_length, dtype=length.dtype, device=length.device)
    return x.unsqueeze(0) < length.unsqueeze(1)


def grad_norm(parameters, norm_type: float = 2.0):
    """
    Calculates norm of parameter gradients

    Args:
        parameters: The list of parameters to clip.
        norm_type: The type of norm to use for clipping.

    """
    if isinstance(parameters, torch.Tensor):
        parameters = [parameters]

    parameters = [p for p in parameters if p.grad is not None]

    if not parameters:
        return 0.0

    return torch.linalg.vector_norm(
        torch.stack([p.grad.norm(norm_type) for p in parameters]),
        ord=norm_type,
    ).item()



"""Common constants and functions for the RVC package."""

from __future__ import annotations

from pathlib import Path

RVC_DIR = Path(__file__).resolve().parent
RVC_CONFIGS_DIR = RVC_DIR / "configs"
RVC_TRAINING_MODELS_DIR = RVC_DIR / "train" / "models"



"""
Module which defines the code for the
"Generate song covers - one-click generation" tab.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from functools import partial

import gradio as gr

from ultimate_rvc.core.generate.song_cover import (
    get_named_song_dirs,
    get_song_cover_name,
    run_pipeline,
)
from ultimate_rvc.core.manage.audio import get_saved_output_audio
from ultimate_rvc.typing_extra import EmbedderModel
from ultimate_rvc.web.common import (
    PROGRESS_BAR,
    exception_harness,
    toggle_intermediate_audio,
    toggle_visibility,
    toggle_visible_component,
    update_dropdowns,
    update_output_name,
    update_value,
)
from ultimate_rvc.web.typing_extra import ConcurrencyId

if TYPE_CHECKING:
    from ultimate_rvc.web.config.main import OneClickSongGenerationConfig, TotalConfig


def render(total_config: TotalConfig, cookiefile: str | None = None) -> None:
    """
    Render "Generate song covers - One-click generation" tab.

    Parameters
    ----------
    total_config : TotalConfig
        Model containing all component configuration settings for the
        Ultimate RVC web UI.
    cookiefile : str, optional
        The path to a file containing cookies to use when downloading
        audio from Youtube.

    """
    with gr.Tab("One-click"):
        tab_config = total_config.song.one_click
        _render_input(tab_config)
        with gr.Accordion("Options", open=False):
            _render_main_options(tab_config)
            _render_conversion_options(tab_config)
            _render_mixing_options(tab_config)
            _render_output_options(tab_config)
            _render_intermediate_audio(tab_config)

        with gr.Row(equal_height=True):
            reset_btn = gr.Button(value="Reset options", scale=2)
            generate_btn = gr.Button("Generate", scale=2, variant="primary")
        song_cover = gr.Audio(
            label="Song cover",
            scale=3,
            waveform_options=gr.WaveformOptions(show_recording_waveform=False),
        )
        song_dirs = total_config.song.multi_step.song_dirs.all
        generate_btn.click(
            partial(
                exception_harness(
                    run_pipeline,
                    info_msg="Song cover generated successfully!",
                ),
                cookiefile=cookiefile,
                progress_bar=PROGRESS_BAR,
            ),
            inputs=[
                tab_config.source.instance,
                tab_config.voice_model.instance,
                tab_config.n_octaves.instance,
                tab_config.n_semitones.instance,
                tab_config.f0_methods.instance,
                tab_config.index_rate.instance,
                tab_config.rms_mix_rate.instance,
                tab_config.protect_rate.instance,
                tab_config.hop_length.instance,
                tab_config.split_voice.instance,
                tab_config.autotune_voice.instance,
                tab_config.autotune_strength.instance,
                tab_config.clean_voice.instance,
                tab_config.clean_strength.instance,
                tab_config.embedder_model.instance,
                tab_config.custom_embedder_model.instance,
                tab_config.sid.instance,
                tab_config.room_size.instance,
                tab_config.wet_level.instance,
                tab_config.dry_level.instance,
                tab_config.damping.instance,
                tab_config.main_gain.instance,
                tab_config.inst_gain.instance,
                tab_config.backup_gain.instance,
                tab_config.output_sr.instance,
                tab_config.output_format.instance,
                tab_config.output_name.instance,
            ],
            outputs=[song_cover, *tab_config.intermediate_audio.all],
            concurrency_limit=1,
            concurrency_id=ConcurrencyId.GPU,
        ).success(
            partial(update_dropdowns, get_named_song_dirs, 3 + len(song_dirs), [], [2]),
            outputs=[
                total_config.song.one_click.cached_song.instance,
                total_config.song.multi_step.cached_song.instance,
                total_config.management.audio.intermediate.instance,
                *song_dirs,
            ],
            show_progress="hidden",
        ).then(
            partial(update_dropdowns, get_saved_output_audio, 1, [], [0]),
            outputs=total_config.management.audio.output.instance,
            show_progress="hidden",
        )
        reset_btn.click(
            lambda: [
                tab_config.n_octaves.value,
                tab_config.n_semitones.value,
                tab_config.f0_methods.value,
                tab_config.index_rate.value,
                tab_config.rms_mix_rate.value,
                tab_config.protect_rate.value,
                tab_config.hop_length.value,
                tab_config.split_voice.value,
                tab_config.autotune_voice.value,
                tab_config.autotune_strength.value,
                tab_config.clean_voice.value,
                tab_config.clean_strength.value,
                tab_config.embedder_model.value,
                tab_config.sid.value,
                tab_config.room_size.value,
                tab_config.wet_level.value,
                tab_config.dry_level.value,
                tab_config.damping.value,
                tab_config.main_gain.value,
                tab_config.inst_gain.value,
                tab_config.backup_gain.value,
                tab_config.output_sr.value,
                tab_config.output_format.value,
                tab_config.show_intermediate_audio.value,
            ],
            outputs=[
                tab_config.n_octaves.instance,
                tab_config.n_semitones.instance,
                tab_config.f0_methods.instance,
                tab_config.index_rate.instance,
                tab_config.rms_mix_rate.instance,
                tab_config.protect_rate.instance,
                tab_config.hop_length.instance,
                tab_config.split_voice.instance,
                tab_config.autotune_voice.instance,
                tab_config.autotune_strength.instance,
                tab_config.clean_voice.instance,
                tab_config.clean_strength.instance,
                tab_config.embedder_model.instance,
                tab_config.sid.instance,
                tab_config.room_size.instance,
                tab_config.wet_level.instance,
                tab_config.dry_level.instance,
                tab_config.damping.instance,
                tab_config.main_gain.instance,
                tab_config.inst_gain.instance,
                tab_config.backup_gain.instance,
                tab_config.output_sr.instance,
                tab_config.output_format.instance,
                tab_config.show_intermediate_audio.instance,
            ],
            show_progress="hidden",
        )


def _render_input(tab_config: OneClickSongGenerationConfig) -> None:
    with gr.Row():
        with gr.Column():
            tab_config.source_type.instantiate()
        with gr.Column():
            tab_config.source.instantiate()
            local_file = gr.Audio(
                label="Source",
                type="filepath",
                visible=False,
                waveform_options=gr.WaveformOptions(show_recording_waveform=False),
            )
            tab_config.cached_song.instance.render()
        tab_config.source_type.instance.input(
            partial(toggle_visible_component, 3),
            inputs=tab_config.source_type.instance,
            outputs=[
                tab_config.source.instance,
                local_file,
                tab_config.cached_song.instance,
            ],
            show_progress="hidden",
        )

        local_file.change(
            update_value,
            inputs=local_file,
            outputs=tab_config.source.instance,
            show_progress="hidden",
        )
        tab_config.cached_song.instance.input(
            update_value,
            inputs=tab_config.cached_song.instance,
            outputs=tab_config.source.instance,
            show_progress="hidden",
        )

    with gr.Row():
        tab_config.voice_model.instance.render()


def _render_main_options(tab_config: OneClickSongGenerationConfig) -> None:
    with gr.Row():
        tab_config.n_octaves.instantiate()
        tab_config.n_semitones.instantiate()


def _render_conversion_options(tab_config: OneClickSongGenerationConfig) -> None:
    with gr.Accordion("Vocal conversion", open=False):
        gr.Markdown("")
        with gr.Accordion("Voice synthesis", open=False):
            with gr.Row():
                tab_config.f0_methods.instantiate()
                tab_config.index_rate.instantiate()
            with gr.Row():
                tab_config.rms_mix_rate.instantiate()
                tab_config.protect_rate.instantiate()
                tab_config.hop_length.instantiate()
        with gr.Accordion("Vocal enrichment", open=False):
            with gr.Row():
                with gr.Column():
                    tab_config.split_voice.instantiate()
                with gr.Column():
                    tab_config.autotune_voice.instantiate()
                    tab_config.autotune_strength.instantiate()
                with gr.Column():
                    tab_config.clean_voice.instantiate()
                    tab_config.clean_strength.instantiate()
            tab_config.autotune_voice.instance.change(
                partial(toggle_visibility, targets={True}),
                inputs=tab_config.autotune_voice.instance,
                outputs=tab_config.autotune_strength.instance,
                show_progress="hidden",
            )
            tab_config.clean_voice.instance.change(
                partial(toggle_visibility, targets={True}),
                inputs=tab_config.clean_voice.instance,
                outputs=tab_config.clean_strength.instance,
                show_progress="hidden",
            )
        with gr.Accordion("Speaker embedding", open=False):
            with gr.Row():
                with gr.Column():
                    tab_config.embedder_model.instantiate()
                    tab_config.custom_embedder_model.instance.render()
                tab_config.sid.instantiate()
            tab_config.embedder_model.instance.change(
                partial(toggle_visibility, targets={EmbedderModel.CUSTOM}),
                inputs=tab_config.embedder_model.instance,
                outputs=tab_config.custom_embedder_model.instance,
                show_progress="hidden",
            )


def _render_mixing_options(tab_config: OneClickSongGenerationConfig) -> None:
    with gr.Accordion("Audio mixing", open=False):
        gr.Markdown("")
        with gr.Accordion("Reverb control on converted vocals", open=False):
            with gr.Row():
                tab_config.room_size.instantiate()
            with gr.Row():
                tab_config.wet_level.instantiate()
                tab_config.dry_level.instantiate()
                tab_config.damping.instantiate()

        with gr.Accordion("Volume controls (dB)", open=False), gr.Row():
            tab_config.main_gain.instantiate()
            tab_config.inst_gain.instantiate()
            tab_config.backup_gain.instantiate()


def _render_output_options(tab_config: OneClickSongGenerationConfig) -> None:
    with gr.Accordion("Audio output", open=False):
        with gr.Row():
            tab_config.output_name.instantiate(
                value=partial(
                    update_output_name,
                    get_song_cover_name,
                    True,  # noqa: FBT003
                ),
                inputs=[
                    gr.State(None),
                    tab_config.cached_song.instance,
                    tab_config.voice_model.instance,
                ],
            )
            tab_config.output_sr.instantiate()
            tab_config.output_format.instantiate()
        with gr.Row():
            tab_config.show_intermediate_audio.instantiate()


def _render_intermediate_audio(tab_config: OneClickSongGenerationConfig) -> None:
    with gr.Accordion(
        "Intermediate audio tracks",
        open=False,
        visible=False,
    ) as intermediate_audio_accordion:
        with gr.Accordion(
            "Step 0: song retrieval",
            open=False,
        ) as song_retrieval_accordion:
            tab_config.intermediate_audio.song.instantiate()
        with (
            gr.Accordion(
                "Step 1a: vocals/instrumentals separation",
                open=False,
            ) as vocals_separation_accordion,
            gr.Row(),
        ):
            tab_config.intermediate_audio.vocals.instantiate()
            tab_config.intermediate_audio.instrumentals.instantiate()
        with (
            gr.Accordion(
                "Step 1b: main vocals/ backup vocals separation",
                open=False,
            ) as main_vocals_separation_accordion,
            gr.Row(),
        ):
            tab_config.intermediate_audio.main_vocals.instantiate()
            tab_config.intermediate_audio.backup_vocals.instantiate()
        with (
            gr.Accordion(
                "Step 1c: main vocals cleanup",
                open=False,
            ) as vocal_cleanup_accordion,
            gr.Row(),
        ):
            tab_config.intermediate_audio.main_vocals_dereverbed.instantiate()
            tab_config.intermediate_audio.main_vocals_reverb.instantiate()
        with gr.Accordion(
            "Step 2: conversion of main vocals",
            open=False,
        ) as vocal_conversion_accordion:
            tab_config.intermediate_audio.converted_vocals.instantiate()
        with gr.Accordion(
            "Step 3: post-processing of converted vocals",
            open=False,
        ) as vocals_postprocessing_accordion:
            tab_config.intermediate_audio.postprocessed_vocals.instantiate()
        with (
            gr.Accordion(
                "Step 4: pitch shift of background tracks",
                open=False,
            ) as pitch_shift_accordion,
            gr.Row(),
        ):
            tab_config.intermediate_audio.instrumentals_shifted.instantiate()
            tab_config.intermediate_audio.backup_vocals_shifted.instantiate()

    tab_config.show_intermediate_audio.instance.change(
        partial(toggle_intermediate_audio, num_components=7),
        inputs=tab_config.show_intermediate_audio.instance,
        outputs=[
            intermediate_audio_accordion,
            song_retrieval_accordion,
            vocals_separation_accordion,
            main_vocals_separation_accordion,
            vocal_cleanup_accordion,
            vocal_conversion_accordion,
            vocals_postprocessing_accordion,
            pitch_shift_accordion,
        ],
        show_progress="hidden",
    )



"""Common utility functions for the core of the Ultimate RVC project."""

from __future__ import annotations

from typing import TYPE_CHECKING, Literal

import lazy_loader as lazy

import hashlib
import json
import shutil
from pathlib import Path

from pydantic import AnyHttpUrl, TypeAdapter, ValidationError

from rich import print as rprint

from ultimate_rvc.common import (
    AUDIO_DIR,
    CUSTOM_EMBEDDER_MODELS_DIR,
    CUSTOM_PRETRAINED_MODELS_DIR,
    MODELS_DIR,
    TRAINING_MODELS_DIR,
    VOICE_MODELS_DIR,
)
from ultimate_rvc.core.exceptions import (
    AudioDirectoryEntity,
    AudioFileEntity,
    Entity,
    HttpUrlError,
    ModelEntity,
    ModelExistsError,
    ModelNotFoundError,
    NotFoundError,
    NotProvidedError,
    UIMessage,
)

if TYPE_CHECKING:
    from collections.abc import Sequence

    import requests

    import gradio as gr

    from ultimate_rvc.typing_extra import Json, StrPath
else:
    requests = lazy.load("requests")


INTERMEDIATE_AUDIO_BASE_DIR = AUDIO_DIR / "intermediate"
SPEECH_DIR = AUDIO_DIR / "speech"
OUTPUT_AUDIO_DIR = AUDIO_DIR / "output"
FLAG_FILE = MODELS_DIR / ".initialized"
TRAINING_AUDIO_DIR = AUDIO_DIR / "training"


def display_progress(
    message: str,
    percentage: float | None = None,
    progress_bar: gr.Progress | None = None,
) -> None:
    """
    Display progress message and percentage in console and potentially
    also Gradio progress bar.

    Parameters
    ----------
    message : str
        Message to display.
    percentage : float, optional
        Percentage to display.
    progress_bar : gr.Progress, optional
        The Gradio progress bar to update.

    """
    rprint(message)
    if progress_bar is not None:
        progress_bar(percentage, desc=message)


def remove_suffix_after(text: str, occurrence: str) -> str:
    """
    Remove suffix after the first occurrence of a substring in a string.

    Parameters
    ----------
    text : str
        The string to remove the suffix from.
    occurrence : str
        The substring to remove the suffix after.

    Returns
    -------
    str
        The string with the suffix removed.

    """
    location = text.rfind(occurrence)
    if location == -1:
        return text
    return text[: location + len(occurrence)]


def copy_files_to_new_dir(files: Sequence[StrPath], directory: StrPath) -> None:
    """
    Copy files to a new directory.

    Parameters
    ----------
    files : Sequence[StrPath]
        Paths to the files to copy.
    directory : StrPath
        Path to the directory to copy the files to.

    Raises
    ------
    NotFoundError
        If a file does not exist.

    """
    dir_path = Path(directory)
    file_paths: list[Path] = []
    for file in files:
        file_path = Path(file)
        if not file_path.exists():
            raise NotFoundError(entity=Entity.FILE, location=file_path)
        file_paths.append(file_path)
    dir_path.mkdir(parents=True)
    for file_path in file_paths:
        shutil.copyfile(file_path, dir_path / file_path.name)


def copy_file_safe(src: StrPath, dest: StrPath) -> Path:
    """
    Copy a file to a new location, appending a number if a file with the
    same name already exists.

    Parameters
    ----------
    src : StrPath
        The source file path.
    dest : StrPath
        The candidate destination file path.

    Returns
    -------
    Path
        The final destination file path.

    """
    dest_path = Path(dest)
    dest_dir = dest_path.parent
    dest_dir.mkdir(parents=True, exist_ok=True)
    dest_file = dest_path
    counter = 1

    while dest_file.exists():
        dest_file = dest_dir / f"{dest_path.stem} ({counter}){dest_path.suffix}"
        counter += 1

    shutil.copyfile(src, dest_file)
    return dest_file


def json_dumps(thing: Json) -> str:
    """
    Dump a JSON-serializable object to a JSON string.

    Parameters
    ----------
    thing : Json
        The JSON-serializable object to dump.

    Returns
    -------
    str
        The JSON string representation of the object.

    """
    return json.dumps(thing, ensure_ascii=False, indent=4)


def json_dump(thing: Json, file: StrPath) -> None:
    """
    Dump a JSON-serializable object to a JSON file.

    Parameters
    ----------
    thing : Json
        The JSON-serializable object to dump.
    file : StrPath
        The path to the JSON file.

    """
    with Path(file).open("w", encoding="utf-8") as fp:
        json.dump(thing, fp, ensure_ascii=False, indent=4)


def json_load(file: StrPath, encoding: str = "utf-8") -> Json:
    """
    Load a JSON-serializable object from a JSON file.

    Parameters
    ----------
    file : StrPath
        The path to the JSON file.
    encoding : str, default='utf-8'
        The encoding of the JSON file.

    Returns
    -------
    Json
        The JSON-serializable object loaded from the JSON file.

    """
    with Path(file).open(encoding=encoding) as fp:
        return json.load(fp)


def get_hash(thing: Json, size: int = 5) -> str:
    """
    Get the hash of a JSON-serializable object.

    Parameters
    ----------
    thing : Json
        The JSON-serializable object to hash.
    size : int, default=5
        The size of the hash in bytes.

    Returns
    -------
    str
        The hash of the JSON-serializable object.

    """
    return hashlib.blake2b(
        json_dumps(thing).encode("utf-8"),
        digest_size=size,
    ).hexdigest()


# NOTE consider increasing size to 16 otherwise we might have problems
# with hash collisions
def get_file_hash(file: StrPath, size: int = 5) -> str:
    """
    Get the hash of a file.

    Parameters
    ----------
    file : StrPath
        The path to the file.
    size : int, default=5
        The size of the hash in bytes.

    Returns
    -------
    str
        The hash of the file.

    """
    with Path(file).open("rb") as fp:
        file_hash = hashlib.file_digest(
            fp,
            lambda: hashlib.blake2b(digest_size=size),  # type: ignore[reportArgumentType]
        )
    return file_hash.hexdigest()


def get_combined_file_hash(files: Sequence[StrPath], size: int = 5) -> str:
    """
    Get the combined hash of a sequence of files.

    Parameters
    ----------
    files : Sequence[StrPath]
        A sequence of paths to the files to hash.
    size : int, default=5
        The size of each hash in bytes.

    Returns
    -------
    str
        The combined hash of the files.

    """
    hasher = hashlib.blake2b(digest_size=size)

    for file in files:
        with Path(file).open("rb") as fp:
            for chunk in iter(lambda: fp.read(4096), b""):
                hasher.update(chunk)

    return hasher.hexdigest()


def get_file_size(url: str) -> int:
    """
    Get the size of a file at a given URL.

    Parameters
    ----------
    url : str
        The URL of the file to get the size of.

    Returns
    -------
    int
        The size of the file at the given URL.

    """
    response = requests.head(url, timeout=10)
    response.raise_for_status()
    return int(response.headers.get("content-length", 0))


def validate_audio_file_exists(
    audio_file: StrPath | None,
    entity: AudioFileEntity,
) -> Path:
    """
    Validate that the provided audio file path is defined and that it
    points to an existing instance of the provided audio file entity.

    Parameters
    ----------
    audio_file : StrPath, optional
        The audio file path to validate.
    entity : AudioFileEntity
        The entity that the audio file path should point to.

    Returns
    -------
    Path
        The validated audio file path.

    Raises
    ------
    NotProvidedError
        If the audio file path is not defined.
    NotFoundError
        If the audio file path does not point to an existing instance
        of the provided audio file entity.

    """
    if not audio_file:
        raise NotProvidedError(entity=entity)
    audio_file_path = Path(audio_file)
    if not audio_file_path.is_file():
        raise NotFoundError(entity=entity, location=audio_file_path)
    return audio_file_path


def validate_audio_dir_exists(
    audio_directory: StrPath | None,
    entity: AudioDirectoryEntity,
) -> Path:
    """
    Validate that the provided audio directory path is defined and that
    it points to an existing instance of the provided audio directory
    entity.

    Parameters
    ----------
    audio_directory : StrPath, optional
        The audio directory path to validate.
    entity : AudioDirectoryEntity
        The entity that the audio directory path should point to.

    Returns
    -------
    Path
        The validated audio directory path.

    Raises
    ------
    NotProvidedError
        If the audio directory path is not defined.
    NotFoundError
        If the audio directory path does not point to an existing
        instance of the provided audio directory entity.

    """
    match entity:
        case Entity.SONG_DIR:
            ui_msg = UIMessage.NO_SONG_DIR
        case _:
            ui_msg = None
    if not audio_directory:
        raise NotProvidedError(entity=entity, ui_msg=ui_msg)
    audio_dir_path = Path(audio_directory)
    if not audio_dir_path.is_dir():
        raise NotFoundError(entity=entity, location=audio_dir_path)
    return audio_dir_path


def validate_model(
    name: str | None,
    entity: ModelEntity,
    mode: Literal["exists", "not_exists"] = "exists",
) -> Path:
    """
    Validate that the provided name is defined and that it identifies
    an existing or non-existing instance of the provided model entity.

    Parameters
    ----------
    name : str | None
        The name of the model to validate.
    entity : ModelEntity
        The entity that the name should identify.
    mode : Literal["exists", "not_exists"]
        The mode of validation. If "exists", the model must exist. If
        "not_exists", the model must not exist.

    Returns
    -------
    Path
        The path to the model entity identified by the provided name.

    Raises
    ------
    NotProvidedError
        If the model name is not defined.
    ModelNotFoundError
        If the model name does not point to an existing instance of the
        provided model entity.
    ModelExistsError
        If the model name points to an existing instance of the
        provided model entity.

    """
    match entity:
        case Entity.VOICE_MODEL:
            ui_msg = UIMessage.NO_VOICE_MODEL
            directory = VOICE_MODELS_DIR
        case Entity.CUSTOM_EMBEDDER_MODEL:
            ui_msg = UIMessage.NO_CUSTOM_EMBEDDER_MODEL
            directory = CUSTOM_EMBEDDER_MODELS_DIR
        case Entity.TRAINING_MODEL:
            ui_msg = None
            directory = TRAINING_MODELS_DIR
        case Entity.CUSTOM_PRETRAINED_MODEL:
            ui_msg = UIMessage.NO_CUSTOM_PRETRAINED_MODEL
            directory = CUSTOM_PRETRAINED_MODELS_DIR
        case Entity.MODEL:
            ui_msg = UIMessage.NO_MODEL
            directory = MODELS_DIR

    if not name:
        raise NotProvidedError(entity=Entity.MODEL_NAME, ui_msg=ui_msg)

    name = name.strip()

    model_path = Path(directory) / name

    if mode == "exists" and not model_path.is_dir():
        raise ModelNotFoundError(entity, name=name)
    if mode == "not_exists" and model_path.is_dir():
        raise ModelExistsError(entity, name=name)

    return model_path


def validate_url(url: str) -> None:
    """
    Validate a HTTP-based URL.

    Parameters
    ----------
    url : str
        The URL to validate.

    Raises
    ------
    NotProvidedError
        If the URL is not provided.
    HttpUrlError
        If the URL is invalid.

    """
    if not url:
        raise NotProvidedError(entity=Entity.URL)
    try:
        TypeAdapter(AnyHttpUrl).validate_python(url)
    except ValidationError:
        raise HttpUrlError(url) from None



import math

import numpy as np

import torch
from torch import nn
from torch.nn import functional as F
from torch.nn.utils import remove_weight_norm
from torch.nn.utils.parametrizations import weight_norm
from torch.utils.checkpoint import checkpoint

from ultimate_rvc.rvc.lib.algorithm.commons import get_padding, init_weights


class ResBlock(nn.Module):
    """
    Residual block with multiple dilated convolutions.

    This block applies a sequence of dilated convolutional layers with Leaky ReLU activation.
    It's designed to capture information at different scales due to the varying dilation rates.

    Args:
        in_channels (int): Number of input channels.
        out_channels (int): Number of output channels.
        kernel_size (int, optional): Kernel size for the convolutional layers. Defaults to 7.
        dilation (tuple[int], optional): Tuple of dilation rates for the convolutional layers. Defaults to (1, 3, 5).
        leaky_relu_slope (float, optional): Slope for the Leaky ReLU activation. Defaults to 0.2.

    """

    def __init__(
        self,
        channels: int,
        kernel_size: int = 7,
        dilation: tuple[int] = (1, 3, 5),
        leaky_relu_slope: float = 0.2,
    ):
        super().__init__()

        self.leaky_relu_slope = leaky_relu_slope

        self.convs1 = nn.ModuleList(
            [
                weight_norm(
                    nn.Conv1d(
                        channels,
                        channels,
                        kernel_size,
                        stride=1,
                        dilation=d,
                        padding=get_padding(kernel_size, d),
                    ),
                )
                for d in dilation
            ],
        )
        self.convs1.apply(init_weights)

        self.convs2 = nn.ModuleList(
            [
                weight_norm(
                    nn.Conv1d(
                        channels,
                        channels,
                        kernel_size,
                        stride=1,
                        dilation=1,
                        padding=get_padding(kernel_size, 1),
                    ),
                )
                for d in dilation
            ],
        )
        self.convs2.apply(init_weights)

    def forward(self, x: torch.Tensor):
        for c1, c2 in zip(self.convs1, self.convs2, strict=False):
            xt = F.leaky_relu(x, self.leaky_relu_slope)
            xt = c1(xt)
            xt = F.leaky_relu(xt, self.leaky_relu_slope)
            xt = c2(xt)
            x = xt + x

        return x

    def remove_weight_norm(self):
        for c1, c2 in zip(self.convs1, self.convs2, strict=False):
            remove_weight_norm(c1)
            remove_weight_norm(c2)


class AdaIN(nn.Module):
    """
    Adaptive Instance Normalization layer.

    This layer applies a scaling factor to the input based on a learnable weight.

    Args:
        channels (int): Number of input channels.
        leaky_relu_slope (float, optional): Slope for the Leaky ReLU activation applied after scaling. Defaults to 0.2.

    """

    def __init__(
        self,
        *,
        channels: int,
        leaky_relu_slope: float = 0.2,
    ):
        super().__init__()

        self.weight = nn.Parameter(torch.ones(channels))
        # safe to use in-place as it is used on a new x+gaussian tensor
        self.activation = nn.LeakyReLU(leaky_relu_slope)

    def forward(self, x: torch.Tensor):
        gaussian = torch.randn_like(x) * self.weight[None, :, None]

        return self.activation(x + gaussian)


class ParallelResBlock(nn.Module):
    """
    Parallel residual block that applies multiple residual blocks with different kernel sizes in parallel.

    Args:
        in_channels (int): Number of input channels.
        out_channels (int): Number of output channels.
        kernel_sizes (tuple[int], optional): Tuple of kernel sizes for the parallel residual blocks. Defaults to (3, 7, 11).
        dilation (tuple[int], optional): Tuple of dilation rates for the convolutional layers within the residual blocks. Defaults to (1, 3, 5).
        leaky_relu_slope (float, optional): Slope for the Leaky ReLU activation. Defaults to 0.2.

    """

    def __init__(
        self,
        *,
        in_channels: int,
        out_channels: int,
        kernel_sizes: tuple[int] = (3, 7, 11),
        dilation: tuple[int] = (1, 3, 5),
        leaky_relu_slope: float = 0.2,
    ):
        super().__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels

        self.input_conv = nn.Conv1d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=7,
            stride=1,
            padding=3,
        )

        self.input_conv.apply(init_weights)

        self.blocks = nn.ModuleList(
            [
                nn.Sequential(
                    AdaIN(channels=out_channels),
                    ResBlock(
                        out_channels,
                        kernel_size=kernel_size,
                        dilation=dilation,
                        leaky_relu_slope=leaky_relu_slope,
                    ),
                    AdaIN(channels=out_channels),
                )
                for kernel_size in kernel_sizes
            ],
        )

    def forward(self, x: torch.Tensor):
        x = self.input_conv(x)
        return torch.stack([block(x) for block in self.blocks], dim=0).mean(dim=0)

    def remove_weight_norm(self):
        remove_weight_norm(self.input_conv)
        for block in self.blocks:
            block[1].remove_weight_norm()


class SineGenerator(nn.Module):
    """
    Definition of sine generator

    Generates sine waveforms with optional harmonics and additive noise.
    Can be used to create harmonic noise source for neural vocoders.

    Args:
        samp_rate (int): Sampling rate in Hz.
        harmonic_num (int): Number of harmonic overtones (default 0).
        sine_amp (float): Amplitude of sine-waveform (default 0.1).
        noise_std (float): Standard deviation of Gaussian noise (default 0.003).
        voiced_threshold (float): F0 threshold for voiced/unvoiced classification (default 0).

    """

    def __init__(
        self,
        samp_rate,
        harmonic_num=0,
        sine_amp=0.1,
        noise_std=0.003,
        voiced_threshold=0,
    ):
        super(SineGenerator, self).__init__()
        self.sine_amp = sine_amp
        self.noise_std = noise_std
        self.harmonic_num = harmonic_num
        self.dim = self.harmonic_num + 1
        self.sampling_rate = samp_rate
        self.voiced_threshold = voiced_threshold

        self.merge = nn.Sequential(
            nn.Linear(self.dim, 1, bias=False),
            nn.Tanh(),
        )

    def _f02uv(self, f0):
        # generate uv signal
        uv = torch.ones_like(f0)
        uv = uv * (f0 > self.voiced_threshold)
        return uv

    def _f02sine(self, f0_values):
        """
        f0_values: (batchsize, length, dim)
        where dim indicates fundamental tone and overtones
        """
        # convert to F0 in rad. The integer part n can be ignored
        # because 2 * np.pi * n doesn't affect phase
        rad_values = (f0_values / self.sampling_rate) % 1

        # initial phase noise (no noise for fundamental component)
        rand_ini = torch.rand(
            f0_values.shape[0],
            f0_values.shape[2],
            device=f0_values.device,
        )
        rand_ini[:, 0] = 0
        rad_values[:, 0, :] = rad_values[:, 0, :] + rand_ini

        # instantanouse phase sine[t] = sin(2*pi \sum_i=1 ^{t} rad)
        tmp_over_one = torch.cumsum(rad_values, 1) % 1
        tmp_over_one_idx = (tmp_over_one[:, 1:, :] - tmp_over_one[:, :-1, :]) < 0
        cumsum_shift = torch.zeros_like(rad_values)
        cumsum_shift[:, 1:, :] = tmp_over_one_idx * -1.0

        sines = torch.sin(torch.cumsum(rad_values + cumsum_shift, dim=1) * 2 * np.pi)

        return sines

    def forward(self, f0):
        with torch.no_grad():
            f0_buf = torch.zeros(f0.shape[0], f0.shape[1], self.dim, device=f0.device)
            # fundamental component
            f0_buf[:, :, 0] = f0[:, :, 0]
            for idx in np.arange(self.harmonic_num):
                f0_buf[:, :, idx + 1] = f0_buf[:, :, 0] * (idx + 2)

            sine_waves = self._f02sine(f0_buf) * self.sine_amp

            uv = self._f02uv(f0)

            noise_amp = uv * self.noise_std + (1 - uv) * self.sine_amp / 3
            noise = noise_amp * torch.randn_like(sine_waves)

            sine_waves = sine_waves * uv + noise

        # merge with grad
        return self.merge(sine_waves)


class RefineGANGenerator(nn.Module):
    """
    RefineGAN generator for audio synthesis.

    This generator uses a combination of downsampling, residual blocks, and parallel residual blocks
    to refine an input mel-spectrogram and fundamental frequency (F0) into an audio waveform.
    It can also incorporate global conditioning.

    Args:
        sample_rate (int, optional): Sampling rate of the audio. Defaults to 44100.
        downsample_rates (tuple[int], optional): Downsampling rates for the downsampling blocks. Defaults to (2, 2, 8, 8).
        upsample_rates (tuple[int], optional): Upsampling rates for the upsampling blocks. Defaults to (8, 8, 2, 2).
        leaky_relu_slope (float, optional): Slope for the Leaky ReLU activation. Defaults to 0.2.
        num_mels (int, optional): Number of mel-frequency bins in the input mel-spectrogram. Defaults to 128.
        start_channels (int, optional): Number of channels in the initial convolutional layer. Defaults to 16.
        gin_channels (int, optional): Number of channels for the global conditioning input. Defaults to 256.
        checkpointing (bool, optional): Whether to use checkpointing for memory efficiency. Defaults to False.

    """

    def __init__(
        self,
        *,
        sample_rate: int = 44100,
        downsample_rates: tuple[int] = (2, 2, 8, 8),  # unused
        upsample_rates: tuple[int] = (8, 8, 2, 2),
        leaky_relu_slope: float = 0.2,
        num_mels: int = 128,
        start_channels: int = 16,  # unused
        gin_channels: int = 256,
        checkpointing: bool = False,
        upsample_initial_channel=512,
    ):
        super().__init__()
        self.upsample_rates = upsample_rates
        self.leaky_relu_slope = leaky_relu_slope
        self.checkpointing = checkpointing

        self.upp = np.prod(upsample_rates)
        self.m_source = SineGenerator(sample_rate)

        # expanded f0 sinegen -> match mel_conv
        self.pre_conv = weight_norm(
            nn.Conv1d(1, upsample_initial_channel // 2, 7, 1, padding=3),
        )

        stride_f0s = [
            math.prod(upsample_rates[i + 1 :]) if i + 1 < len(upsample_rates) else 1
            for i in range(len(upsample_rates))
        ]

        channels = upsample_initial_channel

        self.downsample_blocks = nn.ModuleList([])
        for i, u in enumerate(upsample_rates):
            # handling odd upsampling rates
            stride = stride_f0s[i]
            kernel = 1 if stride == 1 else stride * 2 - stride % 2
            padding = 0 if stride == 1 else (kernel - stride) // 2

            # f0 input gets upscaled to full segment size, then downscaled back to match each upscale step

            self.downsample_blocks.append(
                weight_norm(
                    nn.Conv1d(
                        1,
                        channels // 2 ** (i + 2),
                        kernel,
                        stride,
                        padding=padding,
                    ),
                ),
            )

        self.mel_conv = weight_norm(
            nn.Conv1d(num_mels, channels // 2, 7, 1, padding=3),
        )

        self.mel_conv.apply(init_weights)

        if gin_channels != 0:
            self.cond = nn.Conv1d(256, channels // 2, 1)

        self.upsample_blocks = nn.ModuleList([])
        self.upsample_conv_blocks = nn.ModuleList([])

        for rate in upsample_rates:
            new_channels = channels // 2

            self.upsample_blocks.append(nn.Upsample(scale_factor=rate, mode="linear"))

            self.upsample_conv_blocks.append(
                ParallelResBlock(
                    in_channels=channels + channels // 4,
                    out_channels=new_channels,
                    kernel_sizes=(3, 7, 11),
                    dilation=(1, 3, 5),
                    leaky_relu_slope=leaky_relu_slope,
                ),
            )

            channels = new_channels

        self.conv_post = weight_norm(
            nn.Conv1d(channels, 1, 7, 1, padding=3, bias=False),
        )
        self.conv_post.apply(init_weights)

    def forward(self, mel: torch.Tensor, f0: torch.Tensor, g: torch.Tensor = None):

        f0 = F.interpolate(
            f0.unsqueeze(1),
            size=mel.shape[-1] * self.upp,
            mode="linear",
        )
        har_source = self.m_source(f0.transpose(1, 2)).transpose(1, 2)

        x = self.pre_conv(har_source)
        x = F.interpolate(x, size=mel.shape[-1], mode="linear")
        # expanding spectrogram from 192 to 256 channels
        mel = self.mel_conv(mel)

        if g is not None:
            # adding expanded speaker embedding
            mel = mel + self.cond(g)
        x = torch.cat([mel, x], dim=1)

        for ups, res, down in zip(
            self.upsample_blocks,
            self.upsample_conv_blocks,
            self.downsample_blocks,
            strict=False,
        ):
            x = F.leaky_relu(x, self.leaky_relu_slope)

            if self.training and self.checkpointing:
                x = checkpoint(ups, x, use_reentrant=False)
                x = torch.cat([x, down(har_source)], dim=1)
                x = checkpoint(res, x, use_reentrant=False)
            else:
                x = ups(x)
                x = torch.cat([x, down(har_source)], dim=1)
                x = res(x)

        x = F.leaky_relu(x, self.leaky_relu_slope)
        x = self.conv_post(x)
        x = torch.tanh(x)

        return x

    def remove_weight_norm(self):
        remove_weight_norm(self.pre_conv)
        remove_weight_norm(self.mel_conv)
        remove_weight_norm(self.conv_post)

        for block in self.downsample_blocks:
            block.remove_weight_norm()

        for block in self.upsample_conv_blocks:
            block.remove_weight_norm()



import torch

if torch.cuda.is_available() and torch.cuda.get_device_name().endswith("[ZLUDA]"):

    class STFT:
        def __init__(self):
            self.device = "cuda"
            self.fourier_bases = {}  # Cache for Fourier bases

        def _get_fourier_basis(self, n_fft):
            # Check if the basis for this n_fft is already cached
            if n_fft in self.fourier_bases:
                return self.fourier_bases[n_fft]
            fourier_basis = torch.fft.fft(torch.eye(n_fft, device="cpu")).to(
                self.device,
            )
            # stack separated real and imaginary components and convert to torch tensor
            cutoff = n_fft // 2 + 1
            fourier_basis = torch.cat(
                [fourier_basis.real[:cutoff], fourier_basis.imag[:cutoff]],
                dim=0,
            )
            # cache the tensor and return
            self.fourier_bases[n_fft] = fourier_basis
            return fourier_basis

        def transform(self, input, n_fft, hop_length, window):
            # fetch cached Fourier basis
            fourier_basis = self._get_fourier_basis(n_fft)
            # apply hann window to Fourier basis
            fourier_basis = fourier_basis * window
            # pad input to center with reflect
            pad_amount = n_fft // 2
            input = torch.nn.functional.pad(
                input,
                (pad_amount, pad_amount),
                mode="reflect",
            )
            # separate input into n_fft-sized frames
            input_frames = input.unfold(1, n_fft, hop_length).permute(0, 2, 1)
            # apply fft to each frame
            fourier_transform = torch.matmul(fourier_basis, input_frames)
            cutoff = n_fft // 2 + 1
            return torch.complex(
                fourier_transform[:, :cutoff, :],
                fourier_transform[:, cutoff:, :],
            )

    stft = STFT()
    _torch_stft = torch.stft

    def z_stft(input: torch.Tensor, window: torch.Tensor, *args, **kwargs):
        # only optimizing a specific call from rvc.train.mel_processing.MultiScaleMelSpectrogramLoss
        if (
            kwargs.get("win_length") == None
            and kwargs.get("center") == None
            and kwargs.get("return_complex") == True
        ):
            # use GPU accelerated calculation
            return stft.transform(
                input,
                kwargs.get("n_fft"),
                kwargs.get("hop_length"),
                window,
            )
        # simply do the operation on CPU
        return _torch_stft(
            input=input.cpu(),
            window=window.cpu(),
            *args,
            **kwargs,
        ).to(input.device)

    def z_jit(f, *_, **__):
        f.graph = torch._C.Graph()
        return f

    # hijacks
    torch.stft = z_stft
    torch.jit.script = z_jit
    # disabling unsupported cudnn
    torch.backends.cudnn.enabled = False
    torch.backends.cuda.enable_flash_sdp(False)
    torch.backends.cuda.enable_math_sdp(True)
    torch.backends.cuda.enable_mem_efficient_sdp(False)



"""
Module which exposes functionality for creating and preprocessing
datasets for training voice conversion models.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import lazy_loader as lazy

import shutil
from multiprocessing import cpu_count

from ultimate_rvc.common import TRAINING_MODELS_DIR
from ultimate_rvc.core.common import (
    TRAINING_AUDIO_DIR,
    validate_audio_dir_exists,
    validate_audio_file_exists,
)
from ultimate_rvc.core.exceptions import (
    Entity,
    InvalidAudioFormatError,
    NotProvidedError,
    UIMessage,
)
from ultimate_rvc.typing_extra import AudioExt, AudioSplitMethod, TrainingSampleRate

if TYPE_CHECKING:
    from collections.abc import Sequence
    from pathlib import Path

    import static_ffmpeg

    from ultimate_rvc.typing_extra import StrPath
else:
    static_ffmpeg = lazy.load("static_ffmpeg")


def populate_dataset(name: str, audio_files: Sequence[StrPath]) -> Path:
    """
    Populate the dataset with the provided name with the provided audio
    files.

    If no dataset with the provided name exists, a new dataset with the
    provided name will be created. If any of audio files already exist
    in the dataset, they will be overwritten.

    Parameters
    ----------
    name : str
        The name of the dataset to populate.
    audio_files : list[StrPath]
        The audio files to populate the dataset with.

    Returns
    -------
    The path to the dataset with the provided name.

    Raises
    ------
    NotProvidedError
        If no dataset name or no audio files are provided.

    InvalidAudioFormatError
        If any of the provided audio files are not in a valid format.

    """
    if not name:
        raise NotProvidedError(Entity.DATASET_NAME)

    if not audio_files:
        raise NotProvidedError(Entity.FILES, ui_msg=UIMessage.NO_UPLOADED_FILES)

    static_ffmpeg.add_paths(weak=True)

    import pydub.utils as pydub_utils  # noqa: PLC0415

    audio_paths: list[Path] = []
    for audio_file in audio_files:
        audio_path = validate_audio_file_exists(audio_file, Entity.FILE)
        audio_info = pydub_utils.mediainfo(str(audio_file))
        if not (
            audio_info["format_name"]
            in {
                AudioExt.WAV,
                AudioExt.FLAC,
                AudioExt.MP3,
                AudioExt.OGG,
                AudioExt.AAC,
            }
            or AudioExt.M4A in audio_info["format_name"]
        ):
            raise InvalidAudioFormatError(audio_path, [e.value for e in AudioExt])
        audio_paths.append(audio_path)

    dataset_path = TRAINING_AUDIO_DIR / name.strip()

    dataset_path.mkdir(parents=True, exist_ok=True)

    for audio_path in audio_paths:
        shutil.copyfile(audio_path, dataset_path / audio_path.name)

    return dataset_path


def preprocess_dataset(
    model_name: str,
    dataset: StrPath,
    sample_rate: TrainingSampleRate = TrainingSampleRate.HZ_40K,
    split_method: AudioSplitMethod = AudioSplitMethod.AUTOMATIC,
    chunk_len: float = 3.0,
    overlap_len: float = 0.3,
    filter_audio: bool = True,
    clean_audio: bool = False,
    clean_strength: float = 0.7,
    cpu_cores: int = cpu_count(),
) -> None:
    """
    Preprocess a dataset of audio files for training a voice model.

    Parameters
    ----------
    model_name : str
        The name of the voice model to train. If no voice model
        with the provided name exists for training, a new voice model
        for training will be created with the provided name. If a voice
        model with the provided name already exists for training, then
        its currently associated dataset will be replaced with the
        provided dataset.
    dataset : StrPath
        The path to the dataset to preprocess.
    sample_rate : TrainingSampleRate, default=TrainingSampleRate.HZ_40K
        The target sample rate for the audio files in the provided
        dataset.
    split_method : AudioSplitMethod, default=AudioSplitMethod.AUTOMATIC
        The method to use for splitting the audio files in the provided
        dataset. Use the `Skip` method to skip splitting if the audio
        files are already split. Use the `Simple` method if excessive
        silence has already been removed from the audio files.
        Use the `Automatic` method for automatic silence detection and
        splitting around it.
    chunk_len: float, default=3.0
        length of split audio chunks when using the `Simple` split
        method.
    overlap_len: float, default=0.3
        length of overlap between split audio chunks when using the
        `Simple` split method.
    filter_audio : bool, default=True
        Whether to remove low-frequency sounds from the audio files in
        the provided dataset by applying a high-pass butterworth filter.
    clean_audio : bool, default=False
        Whether to clean the audio files in the provided dataset using
        noise reduction algorithms.
    clean_strength : float, default=0.7
        The intensity of the cleaning to apply to the audio files in the
        provided dataset.
    cpu_cores : int, default=cpu_count()
        The number of CPU cores to use for preprocessing.


    Raises
    ------
    NotProvidedError
        If no model name or dataset is provided.

    """
    if not model_name:
        raise NotProvidedError(Entity.MODEL_NAME)

    dataset_path = validate_audio_dir_exists(dataset, Entity.DATASET)

    model_path = TRAINING_MODELS_DIR / model_name.strip()
    model_path.mkdir(parents=True, exist_ok=True)

    # NOTE The lazy_import function does not work with the package below
    # so we import it here manually
    from ultimate_rvc.rvc.train.preprocess import (  # noqa: PLC0415
        preprocess as train_preprocess,
    )

    train_preprocess.preprocess_training_set(
        str(dataset_path),
        int(sample_rate),
        cpu_cores,
        str(model_path),
        split_method,
        filter_audio,
        clean_audio,
        clean_strength,
        chunk_len,
        overlap_len,
    )



"""Module defining models for representing UI event configurations."""

from __future__ import annotations

from pydantic import BaseModel, ConfigDict, PrivateAttr

from gradio.events import Dependency  # noqa: TC002

from ultimate_rvc.core.exceptions import EventNotInstantiatedError


class ClickEvent(BaseModel):
    """
    Model which represents a click event for a button in the UI.

    Attributes
    ----------
    _instance : Dependency | None
        Internal attribute storing the click event instance.
        Will be null until a click event is assigned manually.
    instance : Dependency
        The click event instance. Attempting to access this field
        before manually assigning a click event will raise an
        EventNotInstantiatedError.

    """

    _instance: Dependency | None = PrivateAttr(default=None)
    model_config = ConfigDict(arbitrary_types_allowed=True)

    @property
    def instance(self) -> Dependency:
        """
        Retrieve the click event instance.

        Returns
        -------
        Dependency
            The click event instance.

        Raises
        ------
        EventNotInstantiatedError
            If the click event instance has not been assigned yet.

        """
        if self._instance is None:
            raise EventNotInstantiatedError
        return self._instance

    @instance.setter
    def instance(self, value: Dependency) -> None:
        self._instance = value


class ManageAudioEventState(BaseModel):
    """
    Event state of the audio management tab.

    Attributes
    ----------
    delete_intermediate_click: ClickEvent
        Click event for deleting intermediate audio files.
    delete_all_intermediate_click: ClickEvent
        Click event for deleting all intermediate audio files.
    delete_speech_click: ClickEvent
        Click event for deleting speech audio files.
    delete_all_speech_click: ClickEvent
        Click event for deleting all speech audio files.
    delete_output_click: ClickEvent
        Click event for deleting output audio files.
    delete_all_output_click: ClickEvent
        Click event for deleting all output audio files.
    delete_dataset_click: ClickEvent
        Click event for deleting dataset audio files.
    delete_all_dataset_click: ClickEvent
        Click event for deleting all dataset audio files.
    delete_all_click: ClickEvent
        Click event for deleting all audio files.

    """

    delete_intermediate_click: ClickEvent = ClickEvent()
    delete_all_intermediate_click: ClickEvent = ClickEvent()
    delete_speech_click: ClickEvent = ClickEvent()
    delete_all_speech_click: ClickEvent = ClickEvent()
    delete_output_click: ClickEvent = ClickEvent()
    delete_all_output_click: ClickEvent = ClickEvent()
    delete_dataset_click: ClickEvent = ClickEvent()
    delete_all_dataset_click: ClickEvent = ClickEvent()
    delete_all_click: ClickEvent = ClickEvent()


class ManageModelEventState(BaseModel):
    """
    Event state of the model management tab.

    Attributes
    ----------
    download_voice_click: ClickEvent
        Click event for downloading voice models.
    upload_voice_click: ClickEvent
        Click event for uploading voice models.
    download_pretrained_click: ClickEvent
        Click event for downloading pretrained models.
    upload_embedder_click: ClickEvent
        Click event for uploading embedder models.
    delete_voice_click: ClickEvent
        Click event for deleting voice models.
    delete_all_voices_click: ClickEvent
        Click event for deleting all voice models.
    delete_embedder_click: ClickEvent
        Click event for deleting embedder models.
    delete_all_embedders_click: ClickEvent
        Click event for deleting all embedder models.
    delete_pretrained_click: ClickEvent
        Click event for deleting pretrained models.
    delete_all_pretraineds_click: ClickEvent
        Click event for deleting all pretrained models.
    delete_trained_click: ClickEvent
        Click event for deleting training models.
    delete_all_trained_click: ClickEvent
        Click event for deleting all training models.
    delete_all_click: ClickEvent
        Click event for deleting all models.

    """

    download_voice_click: ClickEvent = ClickEvent()
    upload_voice_click: ClickEvent = ClickEvent()
    download_pretrained_click: ClickEvent = ClickEvent()
    upload_embedder_click: ClickEvent = ClickEvent()

    delete_voice_click: ClickEvent = ClickEvent()
    delete_all_voices_click: ClickEvent = ClickEvent()
    delete_embedder_click: ClickEvent = ClickEvent()
    delete_all_embedders_click: ClickEvent = ClickEvent()
    delete_pretrained_click: ClickEvent = ClickEvent()
    delete_all_pretraineds_click: ClickEvent = ClickEvent()
    delete_trained_click: ClickEvent = ClickEvent()
    delete_all_trained_click: ClickEvent = ClickEvent()
    delete_all_click: ClickEvent = ClickEvent()



"""
Module which defines extra types used by modules in the
ultimate_rvc.core.manage package.
"""

from __future__ import annotations

from collections.abc import Callable
from enum import StrEnum

from pydantic import BaseModel, RootModel

from ultimate_rvc.typing_extra import PretrainedSampleRate


class VoiceModelTagName(StrEnum):
    """Names of valid voice model tags."""

    ENGLISH = "English"
    JAPANESE = "Japanese"
    OTHER_LANGUAGE = "Other Language"
    ANIME = "Anime"
    VTUBER = "Vtuber"
    REAL_PERSON = "Real person"
    GAME_CHARACTER = "Game character"


class VoiceModelTagMetaData(BaseModel):
    """
    Metadata for a voice model tag.

    Attributes
    ----------
    name : ModelTagName
        The name of the tag.
    description : str
        The description of the tag.

    """

    name: VoiceModelTagName
    description: str


class VoiceModelMetaData(BaseModel):
    """
    Metadata for a voice model.

    Attributes
    ----------
    name : str
        The name of the voice model.
    description : str
        A description of the voice model.
    tags : list[ModelTagName]
        The tags associated with the voice model.
    credit : str
        Who created the voice model.
    added : str
        The date the voice model was created.
    url : str
        An URL pointing to a location where the voice model can be
        downloaded.

    """

    name: str
    description: str
    tags: list[VoiceModelTagName]
    credit: str
    added: str
    url: str


class VoiceModelMetaDataTable(BaseModel):
    """
    Table with metadata for a set of voice models.

    Attributes
    ----------
    tags : list[ModelTagMetaData]
        Metadata for the tags associated with the given set of voice
        models.
    models : list[ModelMetaData]
        Metadata for the given set of voice models.

    """

    tags: list[VoiceModelTagMetaData]
    models: list[VoiceModelMetaData]


VoiceModelMetaDataPredicate = Callable[[VoiceModelMetaData], bool]

VoiceModelMetaDataList = list[list[str | list[VoiceModelTagName]]]


class PretrainedPaths(BaseModel):
    """
    Paths to the generator and discriminator for a pretrained
    model with a given name and sample rate.
    """

    G: str
    D: str


class PretrainedModelMetaData(RootModel[dict[PretrainedSampleRate, PretrainedPaths]]):
    """
    Metadata for a pretrained model with a given name.

    Attributes
    ----------
    root : dict[PretrainedSampleRate, PretrainedPaths]
        Mapping from sample rate to paths to the generator and
        discriminator for the pretrained model with the given name
        at the given sample rate.

    """

    root: dict[PretrainedSampleRate, PretrainedPaths]

    def __getitem__(self, item: PretrainedSampleRate) -> PretrainedPaths:
        """
        Get the paths to the generator and discriminator for the
        pretrained model at the given sample rate.

        Parameters
        ----------
        item : PretrainedSampleRate
            The sample rate for which to get paths to the generator
            and discriminator for the pretrained model.

        Returns
        -------
        PretrainedPaths
            The paths to the generator and discriminator for the
            pretrained model at the given sample rate.

        """
        return self.root[item]

    def keys(self) -> list[PretrainedSampleRate]:
        """
        Get the sample rates for which generator and discriminator
        paths are available for the pretrained model.

        Returns
        -------
        list[PretrainedSampleRate]
            The sample rates for which paths are available for the
            pretrained model.

        """
        return sorted(self.root.keys())


class PretrainedModelMetaDataTable(RootModel[dict[str, PretrainedModelMetaData]]):
    """
    Table with metadata for pretrained models available online.

    Attributes
    ----------
    root : dict[str, PretrainedSampleRates]
        Mapping from the names of pretrained models to metadata for
        those models.

    """

    root: dict[str, PretrainedModelMetaData]

    @property
    def names(self) -> list[str]:
        """
        Get the names of all pretrained models available online.

        Returns
        -------
        list[str]
            The names of all pretrained models available online.

        """
        return self.keys()

    @property
    def default_name(self) -> str | None:
        """
        Get the name of a default pretrained model, if at least one
        pretrained model is available online.

        Returns
        -------
        str | None
            The name of a default pretrained model, or None if no
            pretrained models are available online.

        """
        titan_name = "Titan"
        return titan_name if titan_name in self.names else next(iter(self.names), None)

    @property
    def default_sample_rates(self) -> list[PretrainedSampleRate]:
        """
        Get the sample rates for which instances of the default
        pretrained model are available online.

        Returns
        -------
        list[PretrainedSampleRate]
            The sample rates for which instances of the default
            pretrained model are available online.

        """
        return self.get_sample_rates(self.default_name) if self.default_name else []

    @property
    def default_sample_rate(self) -> PretrainedSampleRate | None:
        """
        Get the first sample rate for which an instance of the default
        pretrained model is available online.

        Returns
        -------
        PretrainedSampleRate | None
            The first sample rate for which an instance of the default
            pretrained model is available online, or if no instances are
            available online.

        """
        return next(iter(self.default_sample_rates), None)

    def get_sample_rates(self, name: str) -> list[PretrainedSampleRate]:
        """
        Get the sample rates for which instances of the pretrained
        model with the provided name are available online.

        Parameters
        ----------
        name : str
            The name of the pretrained model for which to get available
            sample rates.

        Returns
        -------
        list[PretrainedSampleRate]
            The sample rates for which there are instances of the
            pretrained model with the provided name available online.

        """
        return self[name].keys() if name in self.names else []

    def get_sample_rates_with_default(
        self,
        name: str,
    ) -> tuple[PretrainedSampleRate | None, list[PretrainedSampleRate]]:
        """
        Get the sample rates for which instances of the pretrained
        model with the provided name are available online, and the
        default sample rate for that model.

        Parameters
        ----------
        name : str
            The name of the pretrained model for which to get available
            sample rates.

        Returns
        -------
        tuple[PretrainedSampleRate | None, list[PretrainedSampleRate]]
            A tuple containing the default sample rate for the
            pretrained model with the provided name, and a list of all
            sample rates for which there are instances of that model
            available online.

        """
        sample_rates = self.get_sample_rates(name)
        return next(iter(sample_rates), None), sample_rates

    def __getitem__(self, item: str) -> PretrainedModelMetaData:
        """
        Get the metadata for the pretrained model with the given name.

        Parameters
        ----------
        item : str
            The name of the pretrained model for which to get metadata.

        Returns
        -------
        PretrainedSampleRates
            The metadata for the pretrained model with the given name.

        """
        return self.root[item]

    def keys(self) -> list[str]:
        """
        Get the names of all pretrained models available online.

        Returns
        -------
        list[str]
            The names of all pretrained models available online.

        """
        return sorted(self.root.keys())



import os

from ultimate_rvc.common import PRETRAINED_MODELS_DIR


def pretrained_selector(vocoder: str, sample_rate: int) -> tuple[str, str]:
    base_path = os.path.join(PRETRAINED_MODELS_DIR, f"{vocoder.lower()}")

    path_g = os.path.join(base_path, f"f0G{str(sample_rate)[:2]}k.pth")
    path_d = os.path.join(base_path, f"f0D{str(sample_rate)[:2]}k.pth")

    if os.path.exists(path_g) and os.path.exists(path_d):
        return path_g, path_d
    return "", ""



"""
Module which defines the code for the
"Generate song covers - multi-step generation" tab.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from functools import partial

import gradio as gr

from ultimate_rvc.core.generate.common import convert
from ultimate_rvc.core.generate.song_cover import (
    get_named_song_dirs,
    get_song_cover_name,
    mix_song,
    pitch_shift,
    postprocess,
    retrieve_song,
    separate_audio,
)
from ultimate_rvc.core.manage.audio import get_saved_output_audio
from ultimate_rvc.typing_extra import EmbedderModel, RVCContentType
from ultimate_rvc.web.common import (
    exception_harness,
    render_transfer_component,
    setup_transfer_event,
    toggle_visibility,
    toggle_visible_component,
    update_dropdowns,
    update_output_name,
    update_value,
)
from ultimate_rvc.web.typing_extra import ConcurrencyId, SongTransferOption

if TYPE_CHECKING:
    from collections.abc import Sequence

    from ultimate_rvc.web.config.main import MultiStepSongGenerationConfig, TotalConfig


def render(total_config: TotalConfig, cookiefile: str | None = None) -> None:
    """
    Render "Generate song cover - multi-step generation" tab.

    Parameters
    ----------
    total_config : TotalConfig
        Model containing all component configuration settings for the
        Ultimate RVC web UI.
    cookiefile : str, optional
        The path to a file containing cookies to use when downloading
        audio from Youtube.

    """
    tab_config = total_config.song.multi_step
    for input_track in tab_config.input_audio.all:
        input_track.instantiate()
    with gr.Tab("Multi-step"):
        _render_step_0(total_config, cookiefile=cookiefile)
        _render_step_1(tab_config)
        _render_step_2(tab_config)
        _render_step_3(tab_config)
        _render_step_4(tab_config)
        _render_step_5(total_config, tab_config)


def _render_step_0(total_config: TotalConfig, cookiefile: str | None) -> None:
    tab_config = total_config.song.multi_step

    current_song_dir = gr.State(None)
    with gr.Accordion("Step 0: song retrieval", open=True):
        gr.Markdown("")
        with gr.Row():
            with gr.Column():
                tab_config.source_type.instantiate()
            with gr.Column():
                tab_config.source.instantiate()
                local_file = gr.Audio(
                    label="Source",
                    type="filepath",
                    visible=False,
                    waveform_options=gr.WaveformOptions(show_recording_waveform=False),
                )
                tab_config.cached_song.instance.render()

            tab_config.source_type.instance.input(
                partial(toggle_visible_component, 3),
                inputs=tab_config.source_type.instance,
                outputs=[
                    tab_config.source.instance,
                    local_file,
                    tab_config.cached_song.instance,
                ],
                show_progress="hidden",
            )

            local_file.change(
                update_value,
                inputs=local_file,
                outputs=tab_config.source.instance,
                show_progress="hidden",
            )
            tab_config.cached_song.instance.input(
                update_value,
                inputs=tab_config.cached_song.instance,
                outputs=tab_config.source.instance,
                show_progress="hidden",
            )
        with gr.Accordion("Options", open=False):
            song_transfer = _render_song_transfer(
                [SongTransferOption.STEP_1_AUDIO],
                "Song",
            )
        with gr.Row():
            retrieve_song_reset_btn = gr.Button("Reset options")
            retrieve_song_btn = gr.Button("Retrieve song", variant="primary")
        song_transfer_btn = gr.Button("Transfer song")
        song_output = gr.Audio(
            label="Song",
            type="filepath",
            interactive=False,
            waveform_options=gr.WaveformOptions(show_recording_waveform=False),
        )

        retrieve_song_reset_btn.click(
            lambda: gr.Dropdown(value=[SongTransferOption.STEP_1_AUDIO]),
            outputs=song_transfer,
            show_progress="hidden",
        )

        retrieve_song_btn.click(
            partial(
                exception_harness(
                    retrieve_song,
                    info_msg="Song retrieved successfully!",
                ),
                cookiefile=cookiefile,
            ),
            inputs=tab_config.source.instance,
            outputs=[song_output, current_song_dir],
        ).then(
            partial(
                update_dropdowns,
                get_named_song_dirs,
                len(tab_config.song_dirs.all) + 2,
                value_indices=range(len(tab_config.song_dirs.all)),
            ),
            inputs=current_song_dir,
            outputs=[
                *tab_config.song_dirs.all,
                tab_config.cached_song.instance,
                total_config.song.one_click.cached_song.instance,
            ],
            show_progress="hidden",
        ).then(
            partial(update_dropdowns, get_named_song_dirs, 1, [], [0]),
            outputs=total_config.management.audio.intermediate.instance,
            show_progress="hidden",
        )
        setup_transfer_event(
            song_transfer_btn,
            song_transfer,
            song_output,
            tab_config.input_audio.all,
        )


def _render_step_1(tab_config: MultiStepSongGenerationConfig) -> None:
    with gr.Accordion("Step 1: vocal separation", open=False):
        tab_config.input_audio.audio.instance.render()
        tab_config.song_dirs.separate_audio.instance.render()
        with gr.Accordion("Options", open=False):
            with gr.Row():
                tab_config.separation_model.instantiate()
                tab_config.segment_size.instantiate()
            with gr.Row():
                primary_stem_transfer = _render_song_transfer(
                    [SongTransferOption.STEP_2_VOCALS],
                    "Primary stem",
                )
                secondary_stem_transfer = _render_song_transfer(
                    [SongTransferOption.STEP_4_INSTRUMENTALS],
                    "Secondary stem",
                )
        with gr.Row():
            separate_audio_reset_btn = gr.Button("Reset options")
            separate_vocals_btn = gr.Button("Separate vocals", variant="primary")
        with gr.Row():
            primary_stem_transfer_btn = gr.Button("Transfer primary stem")
            secondary_stem_transfer_btn = gr.Button("Transfer secondary stem")

        with gr.Row():
            primary_stem_output = gr.Audio(
                label="Primary stem",
                type="filepath",
                interactive=False,
                waveform_options=gr.WaveformOptions(show_recording_waveform=False),
            )
            secondary_stem_output = gr.Audio(
                label="Secondary stem",
                type="filepath",
                interactive=False,
                waveform_options=gr.WaveformOptions(show_recording_waveform=False),
            )

        separate_audio_reset_btn.click(
            lambda: [
                tab_config.separation_model.value,
                tab_config.segment_size.value,
                gr.Dropdown(value=[SongTransferOption.STEP_2_VOCALS]),
                gr.Dropdown(value=[SongTransferOption.STEP_4_INSTRUMENTALS]),
            ],
            outputs=[
                tab_config.separation_model.instance,
                tab_config.segment_size.instance,
                primary_stem_transfer,
                secondary_stem_transfer,
            ],
            show_progress="hidden",
        )
        separate_vocals_btn.click(
            exception_harness(
                separate_audio,
                info_msg="Vocals separated successfully!",
            ),
            inputs=[
                tab_config.input_audio.audio.instance,
                tab_config.song_dirs.separate_audio.instance,
                tab_config.separation_model.instance,
                tab_config.segment_size.instance,
            ],
            outputs=[primary_stem_output, secondary_stem_output],
            concurrency_limit=1,
            concurrency_id=ConcurrencyId.GPU,
        )
        for btn, transfer, output in [
            (primary_stem_transfer_btn, primary_stem_transfer, primary_stem_output),
            (
                secondary_stem_transfer_btn,
                secondary_stem_transfer,
                secondary_stem_output,
            ),
        ]:
            setup_transfer_event(
                btn,
                transfer,
                output,
                tab_config.input_audio.all,
            )


def _render_step_2(tab_config: MultiStepSongGenerationConfig) -> None:
    with gr.Accordion("Step 2: vocal conversion", open=False):
        tab_config.input_audio.vocals.instance.render()
        tab_config.voice_model.instance.render()
        tab_config.song_dirs.convert_vocals.instance.render()
        with gr.Accordion("Options", open=False):
            with gr.Row():
                tab_config.n_octaves.instantiate()
                tab_config.n_semitones.instantiate()

            converted_vocals_transfer = _render_song_transfer(
                [SongTransferOption.STEP_3_VOCALS],
                "Converted vocals",
            )
            with gr.Accordion("Advanced", open=False):
                with gr.Accordion("Voice synthesis", open=False):
                    with gr.Row():
                        tab_config.f0_methods.instantiate()
                        tab_config.index_rate.instantiate()
                    with gr.Row():
                        tab_config.rms_mix_rate.instantiate()
                        tab_config.protect_rate.instantiate()
                        tab_config.hop_length.instantiate()
                with gr.Accordion("Vocal enrichment", open=False), gr.Row():
                    with gr.Column():
                        tab_config.split_voice.instantiate()
                    with gr.Column():
                        tab_config.autotune_voice.instantiate()
                        tab_config.autotune_strength.instantiate()
                    with gr.Column():
                        tab_config.clean_voice.instantiate()
                        tab_config.clean_strength.instantiate()
                tab_config.autotune_voice.instance.change(
                    partial(toggle_visibility, targets={True}),
                    inputs=tab_config.autotune_voice.instance,
                    outputs=tab_config.autotune_strength.instance,
                    show_progress="hidden",
                )
                tab_config.clean_voice.instance.change(
                    partial(toggle_visibility, targets={True}),
                    inputs=tab_config.clean_voice.instance,
                    outputs=tab_config.clean_strength.instance,
                    show_progress="hidden",
                )
                with gr.Accordion("Speaker embeddings", open=False), gr.Row():
                    with gr.Column():
                        tab_config.embedder_model.instantiate()
                        tab_config.custom_embedder_model.instance.render()
                    tab_config.sid.instantiate()
                tab_config.embedder_model.instance.change(
                    partial(toggle_visibility, targets={EmbedderModel.CUSTOM}),
                    inputs=tab_config.embedder_model.instance,
                    outputs=tab_config.custom_embedder_model.instance,
                    show_progress="hidden",
                )
        with gr.Row():
            convert_vocals_reset_btn = gr.Button("Reset options")
            convert_vocals_btn = gr.Button("Convert vocals", variant="primary")
        converted_vocals_transfer_btn = gr.Button("Transfer converted vocals")
        converted_vocals_track_output = gr.Audio(
            label="Converted vocals",
            type="filepath",
            interactive=False,
            waveform_options=gr.WaveformOptions(show_recording_waveform=False),
        )

        convert_vocals_reset_btn.click(
            lambda: [
                tab_config.n_octaves.value,
                tab_config.n_semitones.value,
                tab_config.f0_methods.value,
                tab_config.index_rate.value,
                tab_config.rms_mix_rate.value,
                tab_config.protect_rate.value,
                tab_config.hop_length.value,
                tab_config.split_voice.value,
                tab_config.autotune_voice.value,
                tab_config.autotune_strength.value,
                tab_config.clean_voice.value,
                tab_config.clean_strength.value,
                tab_config.embedder_model.value,
                tab_config.sid.value,
                gr.Dropdown(value=[SongTransferOption.STEP_3_VOCALS]),
            ],
            outputs=[
                tab_config.n_octaves.instance,
                tab_config.n_semitones.instance,
                tab_config.f0_methods.instance,
                tab_config.index_rate.instance,
                tab_config.rms_mix_rate.instance,
                tab_config.protect_rate.instance,
                tab_config.hop_length.instance,
                tab_config.split_voice.instance,
                tab_config.autotune_voice.instance,
                tab_config.autotune_strength.instance,
                tab_config.clean_voice.instance,
                tab_config.clean_strength.instance,
                tab_config.embedder_model.instance,
                tab_config.sid.instance,
                converted_vocals_transfer,
            ],
            show_progress="hidden",
        )
        convert_vocals_btn.click(
            partial(
                exception_harness(convert, info_msg="Vocals converted successfully!"),
                content_type=RVCContentType.VOCALS,
            ),
            inputs=[
                tab_config.input_audio.vocals.instance,
                tab_config.song_dirs.convert_vocals.instance,
                tab_config.voice_model.instance,
                tab_config.n_octaves.instance,
                tab_config.n_semitones.instance,
                tab_config.f0_methods.instance,
                tab_config.index_rate.instance,
                tab_config.rms_mix_rate.instance,
                tab_config.protect_rate.instance,
                tab_config.hop_length.instance,
                tab_config.split_voice.instance,
                tab_config.autotune_voice.instance,
                tab_config.autotune_strength.instance,
                tab_config.clean_voice.instance,
                tab_config.clean_strength.instance,
                tab_config.embedder_model.instance,
                tab_config.custom_embedder_model.instance,
                tab_config.sid.instance,
            ],
            outputs=converted_vocals_track_output,
            concurrency_id=ConcurrencyId.GPU,
            concurrency_limit=1,
        )
        setup_transfer_event(
            converted_vocals_transfer_btn,
            converted_vocals_transfer,
            converted_vocals_track_output,
            tab_config.input_audio.all,
        )


def _render_step_3(tab_config: MultiStepSongGenerationConfig) -> None:
    with gr.Accordion("Step 3: vocal post-processing", open=False):
        tab_config.input_audio.converted_vocals.instance.render()
        tab_config.song_dirs.postprocess_vocals.instance.render()
        with gr.Accordion("Options", open=False):
            tab_config.room_size.instantiate()
            with gr.Row():
                tab_config.wet_level.instantiate()
                tab_config.dry_level.instantiate()
                tab_config.damping.instantiate()
            effected_vocals_transfer = _render_song_transfer(
                [SongTransferOption.STEP_5_MAIN_VOCALS],
                "Effected vocals",
            )
        with gr.Row():
            postprocess_vocals_reset_btn = gr.Button("Reset options")
            postprocess_vocals_btn = gr.Button("Post-process vocals", variant="primary")
        effected_vocals_transfer_btn = gr.Button("Transfer effected vocals")

        effected_vocals_track_output = gr.Audio(
            label="Effected vocals",
            type="filepath",
            interactive=False,
            waveform_options=gr.WaveformOptions(show_recording_waveform=False),
        )

        postprocess_vocals_reset_btn.click(
            lambda: [
                tab_config.room_size.value,
                tab_config.wet_level.value,
                tab_config.dry_level.value,
                tab_config.damping.value,
                gr.Dropdown(value=[SongTransferOption.STEP_5_MAIN_VOCALS]),
            ],
            outputs=[
                tab_config.room_size.instance,
                tab_config.wet_level.instance,
                tab_config.dry_level.instance,
                tab_config.damping.instance,
                effected_vocals_transfer,
            ],
            show_progress="hidden",
        )
        postprocess_vocals_btn.click(
            exception_harness(
                postprocess,
                info_msg="Vocals post-processed successfully!",
            ),
            inputs=[
                tab_config.input_audio.converted_vocals.instance,
                tab_config.song_dirs.postprocess_vocals.instance,
                tab_config.room_size.instance,
                tab_config.wet_level.instance,
                tab_config.dry_level.instance,
                tab_config.damping.instance,
            ],
            outputs=effected_vocals_track_output,
        )
        setup_transfer_event(
            effected_vocals_transfer_btn,
            effected_vocals_transfer,
            effected_vocals_track_output,
            tab_config.input_audio.all,
        )


def _render_step_4(tab_config: MultiStepSongGenerationConfig) -> None:
    with gr.Accordion("Step 4: pitch shift of background audio", open=False):
        with gr.Row():
            tab_config.input_audio.instrumentals.instance.render()
            tab_config.input_audio.backup_vocals.instance.render()
        with gr.Row():
            tab_config.n_semitones_instrumentals.instantiate()
            tab_config.n_semitones_backup_vocals.instantiate()
        tab_config.song_dirs.pitch_shift_background.instance.render()
        with gr.Accordion("Options", open=False), gr.Row():
            shifted_instrumentals_transfer = _render_song_transfer(
                [SongTransferOption.STEP_5_INSTRUMENTALS],
                "Pitch-shifted instrumentals",
            )
            shifted_backup_vocals_transfer = _render_song_transfer(
                [SongTransferOption.STEP_5_BACKUP_VOCALS],
                "Pitch-shifted backup vocals",
            )
        with gr.Row():
            pitch_shift_instrumentals_btn = gr.Button(
                "Pitch shift instrumentals",
                variant="primary",
            )
            pitch_shift_backup_vocals_btn = gr.Button(
                "Pitch shift backup vocals",
                variant="primary",
            )
        with gr.Row():
            shifted_instrumentals_transfer_btn = gr.Button(
                "Transfer shifted instrumentals",
            )
            shifted_backup_vocals_transfer_btn = gr.Button(
                "Transfer shifted backup vocals",
            )
        pitch_shift_background_reset_btn = gr.Button("Reset options")
        with gr.Row():
            shifted_instrumentals_track_output = gr.Audio(
                label="Pitch-shifted instrumentals",
                type="filepath",
                interactive=False,
                waveform_options=gr.WaveformOptions(show_recording_waveform=False),
            )
            shifted_backup_vocals_track_output = gr.Audio(
                label="Pitch-shifted backup vocals",
                type="filepath",
                interactive=False,
                waveform_options=gr.WaveformOptions(show_recording_waveform=False),
            )

        pitch_shift_background_reset_btn.click(
            lambda: [
                tab_config.n_semitones_instrumentals.value,
                tab_config.n_semitones_backup_vocals.value,
                gr.Dropdown(value=[SongTransferOption.STEP_5_INSTRUMENTALS]),
                gr.Dropdown(value=[SongTransferOption.STEP_5_BACKUP_VOCALS]),
            ],
            outputs=[
                tab_config.n_semitones_instrumentals.instance,
                tab_config.n_semitones_backup_vocals.instance,
                shifted_instrumentals_transfer,
                shifted_backup_vocals_transfer,
            ],
            show_progress="hidden",
        )
        pitch_shift_instrumentals_btn.click(
            exception_harness(
                pitch_shift,
                info_msg="Instrumentals pitch-shifted successfully!",
            ),
            inputs=[
                tab_config.input_audio.instrumentals.instance,
                tab_config.song_dirs.pitch_shift_background.instance,
                tab_config.n_semitones_instrumentals.instance,
            ],
            outputs=shifted_instrumentals_track_output,
        )
        pitch_shift_backup_vocals_btn.click(
            exception_harness(
                pitch_shift,
                info_msg="Backup vocals pitch-shifted successfully!",
            ),
            inputs=[
                tab_config.input_audio.backup_vocals.instance,
                tab_config.song_dirs.pitch_shift_background.instance,
                tab_config.n_semitones_backup_vocals.instance,
            ],
            outputs=shifted_backup_vocals_track_output,
        )
        for btn, transfer, output in [
            (
                shifted_instrumentals_transfer_btn,
                shifted_instrumentals_transfer,
                shifted_instrumentals_track_output,
            ),
            (
                shifted_backup_vocals_transfer_btn,
                shifted_backup_vocals_transfer,
                shifted_backup_vocals_track_output,
            ),
        ]:
            setup_transfer_event(
                btn,
                transfer,
                output,
                tab_config.input_audio.all,
            )


def _render_step_5(
    total_config: TotalConfig,
    tab_config: MultiStepSongGenerationConfig,
) -> None:
    with gr.Accordion("Step 5: song mixing", open=False):
        with gr.Row():
            tab_config.input_audio.main_vocals.instance.render()
            tab_config.input_audio.shifted_instrumentals.instance.render()
            tab_config.input_audio.shifted_backup_vocals.instance.render()
        tab_config.song_dirs.mix.instance.render()
        with gr.Accordion("Options", open=False):
            with gr.Row():
                tab_config.main_gain.instantiate()
                tab_config.inst_gain.instantiate()
                tab_config.backup_gain.instantiate()
            with gr.Row():
                tab_config.output_name.instantiate(
                    value=partial(
                        update_output_name,
                        get_song_cover_name,
                        False,  # noqa: FBT003,
                    ),
                    inputs=[
                        tab_config.input_audio.main_vocals.instance,
                        tab_config.song_dirs.mix.instance,
                    ],
                )
                tab_config.output_sr.instantiate()
                tab_config.output_format.instantiate()
            song_cover_transfer = _render_song_transfer([], "Song cover")
        with gr.Row():
            mix_reset_btn = gr.Button("Reset options")
            mix_btn = gr.Button("Mix song cover", variant="primary")
        song_cover_transfer_btn = gr.Button("Transfer song cover")
        song_cover_output = gr.Audio(
            label="Song cover",
            type="filepath",
            interactive=False,
            waveform_options=gr.WaveformOptions(show_recording_waveform=False),
        )
        mix_reset_btn.click(
            lambda: [
                tab_config.main_gain.value,
                tab_config.inst_gain.value,
                tab_config.backup_gain.value,
                tab_config.output_sr.value,
                tab_config.output_format.value,
                gr.Dropdown(value=[]),
            ],
            outputs=[
                tab_config.main_gain.instance,
                tab_config.inst_gain.instance,
                tab_config.backup_gain.instance,
                tab_config.output_sr.instance,
                tab_config.output_format.instance,
                song_cover_transfer,
            ],
            show_progress="hidden",
        )
        temp_audio_gains = gr.State()
        mix_btn.click(
            partial(
                _pair_audio_tracks_and_gain,
                [
                    tab_config.input_audio.main_vocals.instance,
                    tab_config.input_audio.shifted_instrumentals.instance,
                    tab_config.input_audio.shifted_backup_vocals.instance,
                ],
                [
                    tab_config.main_gain.instance,
                    tab_config.inst_gain.instance,
                    tab_config.backup_gain.instance,
                ],
            ),
            inputs={
                tab_config.input_audio.main_vocals.instance,
                tab_config.input_audio.shifted_instrumentals.instance,
                tab_config.input_audio.shifted_backup_vocals.instance,
                tab_config.main_gain.instance,
                tab_config.inst_gain.instance,
                tab_config.backup_gain.instance,
            },
            outputs=temp_audio_gains,
        ).then(
            exception_harness(mix_song, info_msg="Song cover succesfully generated."),
            inputs=[
                temp_audio_gains,
                tab_config.song_dirs.mix.instance,
                tab_config.output_sr.instance,
                tab_config.output_format.instance,
                tab_config.output_name.instance,
            ],
            outputs=song_cover_output,
        ).then(
            partial(update_dropdowns, get_saved_output_audio, 1, [], [0]),
            outputs=total_config.management.audio.output.instance,
            show_progress="hidden",
        )
        setup_transfer_event(
            song_cover_transfer_btn,
            song_cover_transfer,
            song_cover_output,
            tab_config.input_audio.all,
        )


def _render_song_transfer(
    value: list[SongTransferOption],
    label_prefix: str,
) -> gr.Dropdown:
    return render_transfer_component(value, label_prefix, SongTransferOption)


def _pair_audio_tracks_and_gain(
    audio_components: Sequence[gr.Audio],
    gain_components: Sequence[gr.Slider],
    data: dict[gr.Audio | gr.Slider, Any],
) -> list[tuple[str, int]]:
    """
    Pair audio tracks and gain levels stored in separate gradio
    components.

    This function is meant to first be partially applied to the sequence
    of audio components and the sequence of slider components containing
    the values that should be combined. The resulting function can then
    be called by an event listener whose inputs is a set containing
    those audio and slider components. The `data` parameter in that case
    will contain a mapping from each of those components to the value
    that the component stores.

    Parameters
    ----------
    audio_components : Sequence[gr.Audio]
        Audio components to pair with gain levels.
    gain_components : Sequence[gr.Slider]
        Gain level components to pair with audio tracks.
    data : dict[gr.Audio | gr.Slider, Any]
        Data from the audio and gain components.

    Returns
    -------
    list[tuple[str, int]]
        Paired audio tracks and gain levels.

    Raises
    ------
    ValueError
        If the number of audio tracks and gain levels are not the same.

    """
    audio_tracks = [data[component] for component in audio_components]
    gain_levels = [data[component] for component in gain_components]
    if len(audio_tracks) != len(gain_levels):
        err_msg = "Number of audio tracks and gain levels must be the same."
        raise ValueError(err_msg)
    return [
        (audio_track, gain_level)
        for audio_track, gain_level in zip(audio_tracks, gain_levels, strict=True)
        if audio_track
    ]
